/**
* keysController.js
* Contains handlers for keys related endpoints.
*/
'use strict';
/* eslint-disable no-multi-spaces */
const express          = require('express');
const router           = express.Router();
const changeCase       = require('change-case');
const mem              = require('mem');
const got              = require('got');
const cors             = require('cors');

const config           = require('../config');
const {logger}         = require('../utils/logger');
const {InvalidRequest} = require('../models/error');
/* eslint-enable */

const memGot = mem(got, {maxAge: config.oAuth.metadataCacheMaxAge * 1000});
const getDataFromUrl = url => memGot(url).then(res => JSON.parse(res.body));

const logDebug = message => logger('DEBUG', message);
const logError = message => logger('ERROR', message);
const responseError = (error, res, status) => res.status(status).json(error);
const getJsonError = error => ({
  error: changeCase.snakeCase(error.constructor.name),
  error_description: error.message
});
const responseErrorJson = (error, res, status) =>
  responseError(getJsonError(error), res, status);
const metadataURLForTenantID = tenantId => `https://login.microsoftonline.com/${tenantId}/.well-known/openid-configuration`;

router.get('/azuread', cors(), (req, res) => {
  const tenantId = req.query.tenant_id;
  const getLogMessage = message => `${req.path} : ${message}`;
  const res400 = error => responseErrorJson(error, res, 400);

  logDebug(getLogMessage(`Received request with tenant id: ${tenantId}`));

  if (!tenantId) {
    const error = new InvalidRequest('No tenant id provided.');
    logError(getLogMessage(error.message));
    res400(error);
    return;
  }

  const metadataUrl = metadataURLForTenantID(tenantId);

  getDataFromUrl(metadataUrl)
    .then(metadata => getDataFromUrl(metadata.jwks_uri))
    .then(keys => res.status(200).json(keys))
    .catch(err => {
      const error = new Error(`Could not get the keys. Err: ${err}`);

      logError(getLogMessage(error.message));
      res400(error);
    });
});

module.exports = router;
