/**
* oidcController.js
* Contains the authentication endpoints.
*/
'use strict';

/* eslint-disable no-multi-spaces */
const jwt            = require('jsonwebtoken');
const request        = require('request-promise-native');
const querystring    = require('querystring');
const uuidV4         = require('uuid/v4');
const crypto         = require('crypto');
const {Buffer}       = require('buffer');
const express        = require('express');
const router         = express.Router();

const config         = require('../config');
const { logger }     = require('../utils/logger.js');
const {
  getMetadata,
  getTokenClaims
}                    = require('../models/token');
const {RethrowError} = require('../models/error');
/* eslint-enable */

const Providers = Object.freeze({
  AZURE_V1: 'azure-v1',
  AUTH0: 'auth0',
  ADFS: 'adfs'
});
const logDebug = message => logger('DEBUG', message);
const logError = message => logger('ERROR', message);
const logInfo = message => logger('INFO', message);
const httpOnlyCookie = (res, name, value) => res.cookie(name, value, {httpOnly: true});
const generalGetLogMessage = (reqPath, sessionId, message) => `${reqPath} : ${sessionId} : ${message}`;
const getSessionTokenExpiration = (configOptions, tokenClaims) => {
  if (configOptions.session.token.maxAge) {
    return Math.floor((Date.now() + (configOptions.session.token.maxAge * 1000)) / 1000);
  }

  return tokenClaims.payload.exp;
};

// Login endpoint
function login (req, res) {
  const state = uuidV4();
  const nonce = uuidV4();
  const sessionId = state; // We use the state as auth session id for the logs
  const getLesserLogMessage = message => generalGetLogMessage(req.path, sessionId, message);
  const getLogMessage = message => getLesserLogMessage(`${config.login.parameters.returnTo} ${message}`);
  const resCookieHttpOnly = (name, value) => httpOnlyCookie(res, name, value);

  logDebug(getLesserLogMessage('New authentication request'));

  const returnTo = (req.method === 'POST')
    ? req.body[config.login.parameters.returnTo]
    : req.query[config.login.parameters.returnTo];

  if (!returnTo) {
    logError(getLogMessage('parameter is required'));
    res.status(400).send(`${config.login.parameters.returnTo} parameter is required.`);
    return;
  }

  if (!returnTo.startsWith('/') && !returnTo.startsWith(config.login.allowedUrlStartsWith)) {
    logError(getLogMessage('parameter value is not authorized'));
    res.status(400).send(`${config.login.parameters.returnTo} parameter value is not authorized.`);
    return;
  }

  // Retrieves OIDC provider metadata
  getMetadata()
    .then(metadata => {
      // Set Session cookies
      resCookieHttpOnly(config.session.cookies.returnTo, returnTo);
      resCookieHttpOnly(config.session.cookies.state, state);
      resCookieHttpOnly(config.session.cookies.sessionId, sessionId);
      resCookieHttpOnly(config.session.cookies.nonce, nonce);

      // Initiates an OIDC authorization code grant flow
      // http://openid.net/specs/openid-connect-core-1_0.html#AuthRequest
      const query = querystring.stringify({
        scope: 'openid profile',
        response_type: 'code',
        client_id: config.oAuth.clientId,
        redirect_uri: config.oAuth.callBackUrl,
        response_mode: config.oAuth.responseMode,
        audience: config.oAuth.audience,
        resource: config.oAuth.audience, // Specific to Azure v1
        state,
        nonce
      });
      const authorizationUrl = `${metadata.authorization_endpoint}?${query}`;

      logDebug(getLesserLogMessage(`Login: authorizationUrl: ${authorizationUrl}`));

      // Due to limitations of testing libraries, redirects must be treated differently to get the response correctly.
      /** @see {@link https://github.com/chaijs/chai-http/issues/135} */
      /* istanbul ignore else */
      if (process.env.NODE_ENV === 'test') {
        res.status(200).send(authorizationUrl);
      } else {
        res.redirect(302, authorizationUrl);
      }
    })
    .catch(err => {
      logError(`${req.path} : ${sessionId} : Failed to obtain metadata. Error: ${err}`);
      res.status(400).send(`Failed to obtain metadata. Error: ${err}`);
    });
}

// Handles the authentication response sent by the authorization provider
// http://openid.net/specs/openid-connect-core-1_0.html#AuthResponse
function loginCallback (req, res) {
  const originalState = req.cookies[config.session.cookies.state];
  const sessionId = req.cookies[config.session.cookies.sessionId];
  const responseData = (config.oAuth.responseMode === 'form_post') ? req.body : req.query;
  const {
    code = null,
    state,
    error = null
  } = responseData;
  const sessionState = responseData.session_state;
  const errorDescription = responseData.error_description;
  const getLogMessage = message => generalGetLogMessage(req.path, sessionId, message);
  const resCookieHttpOnly = (name, value) => httpOnlyCookie(res, name, value);
  const handleError = (err, message) => {
    if (err instanceof RethrowError) {
      throw err;
    }

    logError(getLogMessage(message));
    throw new RethrowError({
      message,
      originalError: err
    });
  };

  if (originalState !== state) {
    const message = `Callback called with an invalid state: ${state}`;

    logError(getLogMessage(message));
    res.status(400).send(message);
    return;
  }

  if (error) {
    const message = `Failed to obtain an authorization code. Error: ${error}. ${errorDescription}`;

    logError(getLogMessage(message));
    res.status(400).send(message);
    return;
  }

  let metadata;

  logInfo(getLogMessage(`Callback called with state: ${state}`));
  getMetadata()
    .catch(err => handleError(err, `Failed to obtain metadata. Error: ${err}`))
    .then(data => (metadata = data))
    .then(() =>
      // Request an id_token and an access_token
      // http://openid.net/specs/openid-connect-core-1_0.html#TokenRequest
      request.post({
        url: metadata.token_endpoint,
        headers: {
          'content-type': 'application/x-www-form-urlencoded'
        },
        body: querystring.stringify({
          grant_type: 'authorization_code',
          client_id: config.oAuth.clientId,
          client_secret: config.oAuth.clientSecret,
          code,
          redirect_uri: config.oAuth.callBackUrl,
          resource: config.oAuth.audience // Specific to Azure v1
        })
      })
    )
    .catch(err => handleError(err, `Get token operation failed with error: ${err}`))
    .then(body => {
      // Token response management
      // http://openid.net/specs/openid-connect-core-1_0.html#TokenResponse
      const jsonBody = JSON.parse(body);
      const accessToken = jsonBody.access_token;
      const idToken = jsonBody.id_token;
      const error = jsonBody.error;
      const errorDescription = jsonBody.error_description;

      if (error) {
        logError(getLogMessage(`Get token failed with error: ${error} ${errorDescription}`));
        handleError(error, `Access token rejected with error: ${error} ${errorDescription}`);
      }

      logDebug(getLogMessage(`Get token succeed: id_token: ${idToken} access_token: ${accessToken}`));
      return Promise.all([
        getTokenClaims(req, idToken, config.oAuth.tokenValidation),
        accessToken
      ]);
    })
    .then(([claims, accessToken]) => {
      // Nonce verification
      if (req.cookies[config.session.cookies.nonce] !== claims.payload.nonce) {
        const message = `id_token contains an invalid nonce: ${claims.payload.nonce}`;

        logError(getLogMessage(message));
        handleError(new Error(message), message);
      }

      // Access token validation
      /** @see {@link https://openid.net/specs/openid-connect-core-1_0.html#ImplicitTokenValidation} */
      if (claims.payload.at_hash) {
        const accessTokenHash = crypto
          .createHash('sha256')
          .update(accessToken, 'utf8')
          .digest('hex');
        const leftPart = accessTokenHash.substr(0, accessTokenHash.length / 2);
        const leftPartBase64 = Buffer.from(leftPart, 'hex').toString('base64');

        logDebug(getLogMessage(`at_hash check for value: ${claims.payload.at_hash} against: ${leftPartBase64}`));

        if (leftPartBase64 !== claims.payload.at_hash) {
          const leftPartBase64RemovedEquals = leftPartBase64.replace(/=/g, '');

          logDebug(getLogMessage(`Comparison failed for at_hash. Trying: ${leftPartBase64RemovedEquals}`));

          if (leftPartBase64RemovedEquals !== claims.payload.at_hash) {
            const leftPathBase64RemovedEqualsReplaceMinus = leftPartBase64.replace(/\+/g, '-');

            logDebug(getLogMessage(`Comparison failed for at_hash. Trying: ${leftPathBase64RemovedEqualsReplaceMinus}`));

            if (leftPathBase64RemovedEqualsReplaceMinus !== claims.payload.at_hash) {
              logError(getLogMessage(`Access token not valid after at_hash verification. at_hash: ${claims.payload.at_hash}. encoded access token: ${leftPartBase64}`));
              throw new RethrowError({
                message: 'Invalid access token provided with id token.',
                originalError: new Error('Invalid access token provided with id token.')
              });
            }
          }
        }
      }

      logDebug(getLogMessage('id_token validation succeed'));

      const usernameClaim = claims.payload[config.oAuth.usernameClaim];
      let appToken;

      if (!usernameClaim) {
        logInfo(getLogMessage('The username claim is empty. Your configuration may not specify the correct claim to take.'));
      }

      // Create Application session JWT token
      try {
        appToken = jwt.sign({
          sub: usernameClaim,
          iss: config.session.token.issuer,
          aud: config.session.token.audience,
          exp: getSessionTokenExpiration(config, claims)
        }, config.session.token.signingSecret);
      } catch (err) {
        handleError(err, `Cannot sign a new session token: ${err}`);
      }

      logDebug(getLogMessage(`Application session token created. sub: ${usernameClaim} id_token: ${appToken}`));

      const returnTo = req.cookies[config.session.cookies.returnTo];

      // Set application token session cookie
      resCookieHttpOnly(config.session.cookies.token, appToken);

      // Set session_state cookie if supported by the provider
      // http://openid.net/specs/openid-connect-session-1_0.html
      if (sessionState) {
        resCookieHttpOnly(config.session.cookies.sessionState, sessionState);
      }

      // Remove useless session cookies
      res.clearCookie(config.session.cookies.returnTo);
      res.clearCookie(config.session.cookies.nonce);
      res.clearCookie(config.session.cookies.state);

      /* istanbul ignore else */
      if (process.env.NODE_ENV === 'test') {
        res.status(200).send(returnTo);
      } else {
        res.redirect(returnTo);
      }
    })
    .catch(err => res.status(400).send(err.message));
}

// Initiates a logout
async function logout (req, res) {
  const sessionId = req.cookies[config.session.cookies.sessionId];
  const getLogMessage = message => generalGetLogMessage(req.path, sessionId, message);
  const resCookieHttpOnly = (name, value) => httpOnlyCookie(res, name, value);

  logDebug(getLogMessage('Initiates a logout'));

  const postLogoutReturnTo = req.method === 'POST'
    ? req.body[config.logout.parameters.postLogoutReturnTo]
    : req.query[config.logout.parameters.postLogoutReturnTo];

  if (config.oAuth.provider !== 'adfs') {
    if (!postLogoutReturnTo) {
      const message = `${config.logout.parameters.postLogoutReturnTo} parameter is required`;

      logError(getLogMessage(message));
      res.status(400).send(message);
      return;
    }

    if (!postLogoutReturnTo.startsWith('/') && !postLogoutReturnTo.startsWith(config.logout.allowedUrlStartsWith)) {
      const message = `${config.logout.parameters.postLogoutReturnTo} parameter value is not authorized`;

      logError(getLogMessage(message));
      res.status(400).send(message);
      return;
    }
  }

  try {
    var metadata = await getMetadata();
  } catch (err) {
    const message = `Failed to obtain metadata. Error: ${err}`;

    logError(getLogMessage(message));
    res.status(400).send(message);
    return;
  }

  let logoutUrl;

  // If a new provider have to be added, encapsulate specific functionalities into
  // a conception pattern like strategy and factory.
  switch (config.oAuth.provider) {
    case Providers.AZURE_V1:
      const logoutState = uuidV4();
      const azureQuery = querystring.stringify({
        post_logout_redirect_uri: config.oAuth.logoutReturnToUrl,
        id_token_hint: req.cookies[config.session.cookies.token],
        state: logoutState
      });
      logoutUrl = `${metadata.end_session_endpoint}?${azureQuery}`;

      resCookieHttpOnly(config.session.cookies.logoutState, logoutState);
      break;

    case Providers.ADFS:
      logoutUrl = metadata.end_session_endpoint;

      // Since ADFS doesn't support the redirection of the user to the app
      // after the logout, we immediatly clear the session cookies before redirecting
      // the user to the ADFS logout.
      res.clearCookie(config.session.cookies.token);
      res.clearCookie(config.session.cookies.sessionId);
      res.clearCookie(config.session.cookies.sessionState);
      break;

    default: // Other providers
      const otherQuery = querystring.stringify({
        client_id: config.oAuth.clientId, // For Auth0
        returnTo: config.oAuth.logoutReturnToUrl // For Auth0
      });
      logoutUrl = `${config.oAuth.logoutUrl}?${otherQuery}`; // For Auth0
      break;
  }

  if (postLogoutReturnTo) {
    resCookieHttpOnly(config.session.cookies.postLogoutReturnTo, postLogoutReturnTo);
  }

  logInfo(getLogMessage(`Redirection to: ${logoutUrl}`));

  // istanbul ignore else
  if (process.env.NODE_ENV === 'test') {
    res.status(200).send(logoutUrl);
  } else {
    res.redirect(logoutUrl);
  }
}

// Handles the logout return
function logoutReturn (req, res) {
  const sessionId = req.cookies[config.session.cookies.sessionId];
  const logoutState = req.cookies[config.session.cookies.logoutState];
  const {state} = req.query;
  const getLogMessage = message => generalGetLogMessage(req.path, sessionId, message);

  logInfo(getLogMessage(`Request for post logout from origin: ${req.protocol}://${req.hostname}`));

  // Check logoutState for OpenID Connect session management compliant providers
  if (logoutState && logoutState !== state) {
    const message = `Logout callback called with an invalid state: ${state}`;

    logError(getLogMessage(message));
    res.status(400).send(message);
    return;
  }

  logInfo(getLogMessage('Authentication Provider logout done.'));

  // Cookies clean-up
  res.clearCookie(config.session.cookies.token);
  res.clearCookie(config.session.cookies.sessionId);
  res.clearCookie(config.session.cookies.sessionState);
  res.clearCookie(config.session.cookies.postLogoutReturnTo);
  res.clearCookie(config.session.cookies.logoutState);

  /* istanbul ignore else */
  if (process.env.NODE_ENV === 'test') {
    res.status(200).send(req.cookies[config.session.cookies.postLogoutReturnTo]);
  } else {
    res.redirect(req.cookies[config.session.cookies.postLogoutReturnTo]);
  }
}

// Set Routes
if (config.login.allowedHttpMethods.includes('GET')) {
  router.get('/', login);
}

if (config.login.allowedHttpMethods.includes('POST')) {
  router.post('/', login);
}

if (config.logout.allowedHttpMethods.includes('GET')) {
  router.get('/logout', logout);
}

if (config.logout.allowedHttpMethods.includes('POST')) {
  router.post('/logout', logout);
}

if (config.oAuth.responseMode === 'form_post') {
  router.post('/callback', loginCallback);
} else {
  router.get('/callback', loginCallback);
}

if (config.oAuth.provider !== 'adfs') {
  router.get('/logout/return', logoutReturn);
}

module.exports = router;
