/**
* tokenController.js
* Contains the definition of the token endpoints.
*/
'use strict';

/* eslint-disable no-multi-spaces */
const express    = require('express');
const jwt        = require('jsonwebtoken');
const changeCase = require('change-case');
const router     = express.Router();

const {
  getBearerToken,
  getTokenClaims
}                = require('../models/token');
const config     = require('../config');
const {logger}   = require('../utils/logger');
/* eslint-enable */

class InvalidToken extends Error {}

const getJsonError = (error) => ({
  error: changeCase.snakeCase(error.constructor.name),
  error_description: error.message
});
const responseError = (error, res) => res.status(400).json(error);
const responseToken = (sessionToken, expiresIn, res) => res.status(200).json({
  expires_in: expiresIn,
  session_token: sessionToken
});
const logDebug = message => logger('DEBUG', message);
const logError = message => logger('ERROR', message);
const getSessionTokenExpiration = (configOptions) =>
  Math.floor((Date.now() + (configOptions.session.token.mobileMaxAge * 1000)) / 1000);

router.get('/session_token', (req, res) => {
  const accessToken = getBearerToken(req);
  const resError = error => responseError(error, res);
  const resToken = (token, expiresIn) => responseToken(token, expiresIn, res);

  if (!accessToken) {
    logError('Token is missing or malformed');
    resError(getJsonError(new InvalidToken('Token is missing or malformed')));
    return;
  }

  logDebug(`Verifiying the token claims for: ${accessToken}`);
  getTokenClaims(req, accessToken, config.oAuth.webviewTokenValidation)
    .then(claims => {
      const {payload} = claims;
      const username = payload[config.oAuth.usernameClaim];
      const exp = getSessionTokenExpiration(config);
      const expiresIn = Math.floor(exp - Date.now() / 1000);
      let sessionToken;

      logDebug(`Found username: ${username} for token: ${accessToken}`);

      try {
        sessionToken = jwt.sign({
          sub: username,
          iss: config.session.token.issuer,
          aud: config.session.token.audience,
          exp
        }, config.session.token.signingSecret);
      } catch (err) {
        logError('Could not sign a new session token.');
        throw err;
      }

      logDebug(`New session token generated: ${sessionToken}`);
      resToken(sessionToken, expiresIn);
    })
    .catch(err => {
      logError(`An error occured: ${err}`);
      resError(getJsonError(new InvalidToken('Token is invalid')));
    });
});

module.exports = router;
