/**
 * context.js
 * Context management functions
 */
'use strict';

const config = require('../../config');
const {
  getTokenClaims,
  getBearerToken
} = require('./token.js');

/**
 * Authenticate the request
 * @param {Object} request A request.
 * @return {Promise} A promise.
 */
function authenticate (request) {
  // Retrieve the authenticated username from iisnode promoted server variables
  // Authentication methods:
  // - Basic
  // - Windows
  // - .Net Http authentication module setting the user principal.
  // Security: in IIS, the application authentication MUST have anonymous access disabled
  // or an Http module is configured to define the user principal.
  if (request.headers['x-iisnode-auth_user']) {
    const userName = request.headers['x-iisnode-auth_user'];
    request.userName = userName;
    request.impersonateUserName = getImpersonateUserName(request);
    return Promise.resolve(userName);
  }

  // Authorization with a Bearer token
  if (request.headers.authorization) {
    const token = getBearerToken(request);

    if (!token) {
      return Promise.reject(new Error('Only a Bearer token is supported in the authorization header'));
    }

    return getTokenClaims(request, token, config.oAuth.jwksClient)
      .then(token => {
        // The API can receive two types of access_token:
        // - one generated by an client-credentials flow (dameons)
        // - one generated by an interactive flow (a user)
        const isClientCredentials =
          (config.oAuth.accessTokenProvider === 'Auth0' && token.payload.gty === 'client-credentials') || (config.oAuth.accessTokenProvider === 'Azure' && token.payload.scp !== 'user_impersonation');

        if (isClientCredentials) {
          request.appClientId = token.payload[config.oAuth.accessTokenAppIdClaim];
        } else {
          request.userName = token.payload[config.oAuth.accessTokenUsernameClaim];
          request.impersonateUserName = getImpersonateUserName(request);
        }

        return token.payload.sub;
      });
  }

  // Use env variable in development only
  if ((process.env.NODE_ENV === 'development') && process.env.AUTH_USER) {
    request.userName = process.env.AUTH_USER;
    return Promise.resolve(process.env.AUTH_USER);
  }

  return Promise.reject(new Error('No valid authentication information found'));
}

/**
 * Get the impersonate username.
 * @param {Object} request A request.
 * @return {String} The username.
 */
function getImpersonateUserName (request) {
  const impersonateUserNameHeader = process.env.GraphqlImpersonateUserNameHttpHeader || 'x-wfgen-impersonate-username';
  const userName = request.headers[impersonateUserNameHeader] || process.env.IMPERSONATE_USER;
  return userName ? userName.toLowerCase() : null;
}

/**
 * Creates a workflow context.
 * @param {Object} request A request.
 * @return {Object} The workflow context.
 */
function getWorkflowContext (request) {
  let workflowContext = {
    operationCount: 0,
    operationErrorCount: 0
  };

  if (request.userName) {
    workflowContext['authUserName'] = request.userName;
  }
  if (request.impersonateUserName) {
    workflowContext['impersonateUserName'] = request.impersonateUserName;
  }
  if (request.appClientId) {
    workflowContext['appClientId'] = request.appClientId;
  }

  return workflowContext;
}

module.exports = {
  authenticate,
  getImpersonateUserName,
  getWorkflowContext
};
