/**
* token.js
* Contains the JWT Token management.
*/
'use strict';

/* eslint-disable no-multi-spaces */
const jwt              = require('jsonwebtoken');
const jwksClient       = require('jwks-rsa');
const mem              = require('mem');
const got              = require('got');

const config           = require('../../config');
const { logger }       = require('../../utils/logger.js');
const { RethrowError } = require('./error');
/* eslint-enable */

// OIDC provider metadata
const getMetadata = mem(
  () => got(config.oAuth.metadataUrl)
    .then(result => JSON.parse(result.body)),
  {maxAge: config.oAuth.metadataCacheMaxAge * 1000}
);
const logError = message => logger('ERROR', message);

/**
 * Get a bearer token
 * @param {Object} req The request.
 * @return {String} The token.
 */
function getBearerToken (req) {
  const {authorization} = req.headers;

  if (!authorization) {
    return null;
  }

  const parts = authorization.split(' ');
  const [tokenType, token] = parts;

  if (!(parts.length === 2 && tokenType === 'Bearer' && token)) {
    return null;
  }

  return token;
}

/**
 * Get the JWT claims
 * @param {Object} req The request
 * @param {String} token The token
 * @param {Object} tokenValidationOptions Options for the validation of the token.
 * @return {Object} The claims.
 */
function getTokenClaims (req, token, tokenValidationOptions) {
  if (!token) {
    return Promise.reject(new Error('A token is required'));
  }

  const getLogMessage = message => `${req.path} : ${message}`;
  const handleError = (err, message) => {
    if (err instanceof RethrowError) {
      throw err;
    }

    logError(getLogMessage(message));
    throw new RethrowError({
      message,
      originalError: err
    });
  };
  let tokenClaims;

  return getMetadata()
    .then(metadata => [
      jwt.decode(token, {complete: true}),
      Object.assign({
        jwksUri: metadata.jwks_uri
      }, config.oAuth.jwksClient)
    ])
    .catch(err => handleError(err, `Invalid bearer token. Decoding error : ${err}`))
    .then(([claims, jwksClientOptions]) => {
      tokenClaims = claims;
      return getSigningKey(claims.header.kid || claims.header.x5t, jwksClientOptions);
    })
    .catch(err => handleError(err, `Token validation error. Failed to get the signing key with error : ${err}`))
    .then(key => jwt.verify(token, key, tokenValidationOptions))
    .catch(err => handleError(err, `Token validation error : ${err}`))
    .then(() => tokenClaims);
}

/**
 * Get the signing key for the received token.
 * @param {String} kid The id of the public key.
 * @param {Object} jwksClientOptions Options for jwks client.
 * @returns {Promise<String>} A promise resolving a string representing the key.
 */
function getSigningKey (kid, jwksClientOptions) {
  return new Promise((resolve, reject) => {
    const client = jwksClient(jwksClientOptions);
    const getKeys = client.getKeys.bind(client);

    // Workaround proxy to get the certificate because jwks-rsa doesn't support getting
    // the key when there's only an x5t value and no kid yet.
    client.getKeys = function (cb) {
      getKeys((err, keys) => {
        if (err) {
          cb(err, keys);
          return;
        }

        const proxiedKeys = keys
          .map(k => {
            if (!k.kid && k.x5t) {
              k.kid = k.x5t;
            }

            return k;
          });

        cb(err, proxiedKeys);
      });
    };

    client.getSigningKey(kid, (err, key) => {
      if (err) {
        reject(err);
        return;
      }

      resolve(key.publicKey || key.rsaPublicKey);
    });
  });
}

module.exports = {
  getTokenClaims,
  getBearerToken,
  getMetadata
};
