/**
 * index.js
 * Data loaders
 */
'use strict';
const DataLoader = require('dataloader');
const {getBatchUsersByUserName} = require('./user');
const {getBatchCategoryByName} = require('./category');
const {getWorkflowNodesByOperation, getBatchWorkflowNodes} = require('./workflownode');

/**
 * Creates Data Loaders.
 * @param {Object} context A Context.
 * @return {Array} An array of DataLoaders.
 */
function createLoaders (config, context) {
  return {
    operations: new DataLoader(
      keys => getWorkflowNodesByOperation(context, keys),
      { cacheKeyFn: cacheKey,
        cache: config.cacheEnabled,
        batch: config.batchEnabled,
        maxBatchSize: config.maxBatchSize }
    ),
    users: new DataLoader(
      keys => getBatchWorkflowNodes('getUsers', context, keys),
      { cache: config.cacheEnabled,
        batch: config.batchEnabled,
        maxBatchSize: config.maxBatchSize }
    ),
    usersByUserName: new DataLoader(
      keys => getBatchUsersByUserName(context, keys),
      { cache: config.cacheEnabled,
        batch: config.batchEnabled,
        maxBatchSize: config.maxBatchSize }
    ),
    processes: new DataLoader(
      keys => getBatchWorkflowNodes('getProcesses', context, keys),
      { cache: config.cacheEnabled,
        batch: config.batchEnabled,
        maxBatchSize: config.maxBatchSize }
    ),
    categories: new DataLoader(
      keys => getBatchWorkflowNodes('getCategories', context, keys),
      { cache: config.cacheEnabled,
        batch: config.batchEnabled,
        maxBatchSize: config.maxBatchSize }
    ),
    categoryByName: new DataLoader(
      keys => getBatchCategoryByName(context, keys),
      { cache: config.cacheEnabled,
        batch: config.batchEnabled,
        maxBatchSize: config.maxBatchSize }
    ),
    participants: new DataLoader(
      keys => getBatchWorkflowNodes('getParticipants', context, keys),
      { cache: config.cacheEnabled,
        batch: config.batchEnabled,
        maxBatchSize: config.maxBatchSize }
    )
  };
}

/**
 * Clear all Loaders.
 * @param {Object} loaders Loaders
 */
function clearAllLoaders (loaders) {
  Object.keys(loaders).map(key => loaders[key].clearAll());
}

/**
* Creates a cache key.
* @param {Object} key A key.
* @return {String} A serialized representation of the key.
*/
function cacheKey (key) {
  var result;
  if (typeof key === 'object') {
    result = JSON.stringify(key);
  } else {
    result = String(key);
  }
  return result;
}

module.exports = {createLoaders, clearAllLoaders};
