/**
 * schema.js
 * GraphQL Schema
 */
'use strict';
const fs = require('fs');
const path = require('path');
const merge = require('lodash.merge');
const { makeExecutableSchema, addMockFunctionsToSchema } = require('graphql-tools');

/**
 * @description Creates a GraphQL Schema.
 * @param {Object} config A configuration.
 * @return {Object} The schema.
 */
function createSchema (config) {
  // Load type definitions
  const typeDefs = [];
  fs.readdirSync(path.join(__dirname, 'typedefs'))
    .forEach(file => typeDefs.push(require('./typedefs/' + file)));

  // Load resolvers
  var resolvers = {};
  fs.readdirSync(path.join(__dirname, 'resolvers'))
    .forEach(file => {
      const functions = require('./resolvers/' + file);
      merge(resolvers, functions);
    });

  // Creates the schema
  const schema = makeExecutableSchema({
    typeDefs: typeDefs,
    resolvers: resolvers
  });

  // Load mocks if needed
  if (config.mocksEnabled) {
    var mocks = {};

    fs.readdirSync(path.join(__dirname, 'mocks'))
      .forEach(file => {
        const functions = require('./mocks/' + file);
        merge(mocks, functions);
      });

    // Add mocks if needed, modifies schema in place
    addMockFunctionsToSchema({
      schema,
      preserveResolvers: true,
      mocks: mocks
    });
  }

  return schema;
}

module.exports = { createSchema };
