/**
 * action.js
 * GraphQL Type definitions using schema language
 */
'use strict';

const actionTypeDefs = `

# An action.
type Action implements Node {

  # Action identifier.
  id: ID!

  # Action name.
  name: String!

  # Action description.
  description: String!

  # Action number.
  number: Int!

  # Request associated to the action.
  request: Request
 
  # Action status.
  status: ActionStatus!

  # Action sub-status.
  subStatus: ActionSubStatus!

  # Activity associated to the action.
  activity: Activity!

  # Date and time limit of the action.
  limit: DateTime

  # Child request associated to the action.
  childRequest: Request

  # User who initiated the action.
  initiator: User
  
  # User assigned to the action.
  assignee: User

  # Delegated user assigned to the action.
  delegatedAssignee: User

  # User who assigned the action.
  assignor: User

  # Date and time when the action was assigned.
  assignedAt: DateTime

  # Date and time when the action was aborted.
  abortedBy: User
  
  # Date and time when the action was opened. 
  openedAt: DateTime

  # Date and time when the action was started.
  startedAt: DateTime

  # Date and time when the action was updated. 
  updatedAt: DateTime

  # Date and time when the action was closed. 
  closedAt: DateTime
  
  # URL of the action follow-up form.
  followUpFormUrl: URI!

  # URL of the action follow-up form displayed in the portal.
  followUpFormPortalUrl: URI!

  # URL to launch the action.
  launchUrl: URI

  # URL to launch the action in the portal.
  launchPortalUrl: URI

  # Process data associated to the action.
  dataset (page: Page, orderBy: ActionDataListOrder, filter: ActionDataFilter) : ActionDataListPage!

  # Parameters required to complete the action.
  parameters(filter: ActionParameterFilter): ActionParameterListPage

  # Indicates if the user has permissions to assign the action.
  canAssign: Boolean

  # Indicates if the user has permissions to assign the action to a specific user.
  canAssignUser(assigneeId: ID!): Boolean

  # Indicates if the user has permissions to cancel the action.
  canCancel: Boolean

  # Indicates if the user has permissions to cancel the existing action assignment.
  canCancelAssignment: Boolean

  # Indicates if the user has permissions to complete the action.
  canComplete: Boolean

  # List of users to whom the action can be assigned.
  assignableUsers(page: Page, filter: ActionAssignableUsersFilter): UserListPage!
}

# Represent the possible action statuses.
enum ActionStatus {

  # An action that is still open.
  OPEN

  # An action that has been closed.
  CLOSED 
}

# Represent the possible action sub-statuses.
enum ActionSubStatus {

  # An action that hasn't been started.
  NOT_RUNNING_NOT_STARTED

  # An action that is still running.
  RUNNING

  # An action that has been completed.
  COMPLETED

  # An action that has been aborted.
  ABORTED
}

# Represent the possible action roles.
enum ActionRole {

  # An user who has been assigned to an action.
  ASSIGNEE

  # An user who has assigned someone to an action.
  ASSIGNOR

  # An user who is a supervisor of an action.
  SUPERVISOR 
}

# A paginated list of actions.
type ActionListPage {

  # Identifies the total count of items in the page.
  totalCount: Int!

  # Indicates if there are more pages to fetch.
  hasNextPage: Boolean!

  # Indicates if there are any pages prior to the current page.
  hasPreviousPage: Boolean!

  # A list of actions.
  items: [Action]
}

# A filter on actions.
input ActionFilter {

  # Filter actions by process name.
  processName: String
  
  # Filter actions by process version.
  processVersion: Int

  # Filter actions by request identifier.
  requestId: ID

  # Filter actions by request number.
  requestNumber: Int

  # Filter actions by status.
  status: ActionStatus

  # Filter actions by a list of sub-status.
  subStatuses: [ActionSubStatus]

  # Indicates if actions are in test.
  isTest: Boolean

  # Indicates if actions are or were late.
  isDueDateExceeded: Boolean

  # Indicates if actions are assigned.
  isAssigned: Boolean

  # Indicates if actions are assigned.
  isSelfService: Boolean

  # Filter actions by role.
  as: ActionRole

  # Filter actions by assignee identifier.
  assigneeId: ID

  # Filter actions by participant identifier.
  participantId: ID

  # Filter actions since specified creation date.
  createdSince: DateTime

  # Filter actions since specified assignment date.
  assignedSince: DateTime
}

# A filter on actions associated to a request.
input RequestActionFilter {

  # Filter actions by status.
  status: ActionStatus

  # Indicates if actions are in test.
  isTest: Boolean
  
  # Indicates if actions are or were late.
  isDueDateExceeded: Boolean

  # Indicates if actions are assigned.
  isAssigned: Boolean

  # Indicates if actions are self-service.
  isSelfService: Boolean

  # Filter actions by role.
  as: ActionRole

  # Indicates if the action is displayed in the request's actions history.
  isVisibleInHistory: Boolean

}

# An action list order.
input ActionListOrder {

  # Field to order actions by.
  field: ActionListOrderField

  # Ordering direction.
  direction: ListOrderDirection
}

# A list of fields to order a list of actions.
enum ActionListOrderField {

  # Order actions by number.
  NUMBER

  # Order actions by when they were opened.
  OPENED_AT

  # Order actions by when they were closed.
  CLOSED_AT

  # Order actions by date limit.
  LIMIT

  # Order actions by status.
  STATUS
}

# A complete action input.
input CompleteActionInput {

  # Client mutation identifier.
  clientMutationId: String

  # Action identifier.
  id: ID

  # Request number.
  requestNumber: Int

  # Action number.
  number: Int

  # User identifier that is used by the user to perform the mutation on behalf of another person.
  onBehalfOf: ID

  # Array of parameters.
  parameters: [ParameterInput]
}

# A complete action result.
type CompleteActionResult {

  # Client mutation identifier.
  clientMutationId: String

  # The action that has been completed.
  action: Action!
}

# An assign action input.
input AssignActionInput {

  # Client mutation identifier.
  clientMutationId: String

  # Action identifier.
  id: ID

  # Request number.
  requestNumber: Int

  # Action number.
  number: Int

  # User identifier of the person assigned to the action.
  assigneeId: ID!

  # User identifier that is used by the user to perform the mutation on behalf of another person.
  onBehalfOf: ID
}

# An assign action result.
type AssignActionResult {

  # Identifies the client mutation identifier.
  clientMutationId: String

  # The action that has been assigned.
  action: Action!
}

# A cancel action input.
input CancelActionInput {

  # Client mutation identifier.
  clientMutationId: String

  # Action identifier.
  id: ID

  # Request number.
  requestNumber: Int

  # Action number.
  number: Int

  # Cancellation type.
  type: ActionCancellationType

  # Exception.
  exception: ExceptionInput

  # User identifier that is used by the user to perform the mutation on behalf of another person.
  onBehalfOf: ID
}

# A cancel action result.
type CancelActionResult {

  # Client mutation identifier.
  clientMutationId: String

  # The action that has been cancelled.
  action: Action!
}

# A cancel a request's actions by name input.
input CancelRequestActionsByNameInput {

  # Client mutation identifier.
  clientMutationId: String

  # Request number.
  requestNumber: Int

  # Activity name.
  activityName: String!

  # Cancellation type.
  type: ActionCancellationType

  # Exception.
  exception: ExceptionInput

  # User identifier that is used by the user to perform the mutation on behalf of another person.
  onBehalfOf: ID
}

# A cancel a request's actions by name result.
type CancelRequestActionsByNameResult {

  # Client mutation identifier.
  clientMutationId: String

  # List of actions that have been cancelled.
  actions: [Action]!
}

# A cancel action assignment input.
input CancelActionAssignmentInput {

  # Client mutation identifier.
  clientMutationId: String

  # Action identifier.
  id: ID

  # Request number.
  requestNumber: Int

  # Action number.
  number: Int

  # User identifier that is used by the user to perform the mutation on behalf of another person.
  onBehalfOf: ID
}

# A cancel action assignment result.
type CancelActionAssignmentResult {

  # Client mutation identifier.
  clientMutationId: String

  # The action that has been updated.
  action: Action!
}

# An exception input.
input ExceptionInput {

  # Exception message.
  message: String!

  # Exception source.
  source: String
}

# A list of possible action cancellation exceptions
enum ActionCancellationType {

  # Cancellation exception.
  CANCEL,

  # Error exception.
  ERROR,
}

# A filter on action assignable users.
input ActionAssignableUsersFilter {

  # Filter by user's username, first name and last name.
  nameContains: String

}
`;

module.exports = actionTypeDefs;
