/**
 * application.js
 * GraphQL Type definitions using schema language
 */
'use strict';

const applicationTypeDefs = `

# An application.
type Application implements Node {

  # Application identifier.
  id: ID!

  # Application name.
  name: String!

  # Application description.
  description: String!

  # Application type.
  type: ApplicationType!

  # Application Url.
  url: String

  # Application Wsdl.
  wsdl: String

  # Application username.
  username: String

  # Application password.
  password: String

  # Application assembly name.
  assemblyName: String

  # Application assembly class name.
  assemblyClassName: String

  # Application method.
  method: String

  # Application content type.
  contentType: String

  # Application context format.
  contextFormat: ApplicationContextFormatType

  # Application schema definition.
  isSchemaDefinitionEmbedded: Boolean

  # Application parameters.
  parameters(orderBy: ApplicationParameterListOrder) : ApplicationParameterListPage

  # Application default.
  isDefault: Boolean!

  # Application active.
  isActive: Boolean!

  # Application client identifier.
  clientId: String
}

# Represent the application types.
enum ApplicationType {

 # Application of type Assembly.
 ASSEMBLY

 # Application of type Asynchronous web procedure.
 WEBPROCASYNC

 # Application of type Incoming webhook.
 INWEBHOOK

 # Application of type SOAP web service.
 WEBSERVICE

 # Application of type WCF service.
 WCFSERVICE

 # Application of type Web application.
 WEBAPP

 # Application of type Web procedure.
 WEBPROC

 # Application of type Non interactive client.
 NONINTERACTIVECLIENT
}

# Represent the application context format types.
enum ApplicationContextFormatType {

  # Application JSON context format type.
  JSON

  # Application XML ADO.NET DataSet context format type.
  WFCONTEXT

  # Application XML ADO Recordset context format type
  RECORDSET
}

# An application parameter.
type ApplicationParameter implements Node {

  # Application parameter identifier.
  id: ID!

  # Application parameter name.
  name: String!

  # Application parameter description.
  description: String!

  # Application parameter data type.
  dataType: WorkflowDataType!

  # Application parameter direction.
  direction: ParameterDirection!

  # Application parameter required.
  isRequired: Boolean!

  # Application parameter default.
  isDefault: Boolean!
}

# A paginated list of application parameters.
type ApplicationParameterListPage {

  # Identifies the total count of items in the page.
  totalCount: Int!

  # Indicates if there are more pages to fetch.
  hasNextPage: Boolean!

  # Indicates if there are any pages prior to the current page.
  hasPreviousPage: Boolean!

  # List of application parameters.
  items: [ApplicationParameter]
}

# A paginated list of applications.
type ApplicationListPage {

  # Identifies the total count of items in the page.
  totalCount: Int!

  # Indicates if there are more pages to fetch.
  hasNextPage: Boolean!

  # Indicates if there are any pages prior to the current page.
  hasPreviousPage: Boolean!

  # List of applications.
  items: [Application]
}

# A filter on application.
input ApplicationFilter {
    
  # Filter application by name.  
  nameContains: String

  # Filter application by description.  
  descriptionContains: String

  # Filter application by type.  
  type: ApplicationType

  # Indicate if the application is active.  
  isActive: Boolean
}

# An application list order.
input ApplicationListOrder {

  # Field to order application by.
  field: ApplicationListOrderField

  # Ordering direction.
  direction: ListOrderDirection
}

# A list of fields to order an application list.
enum ApplicationListOrderField {

  # Order applications by type.
  TYPE

  # Order applications by display order.
  DISP_ORDER

  # Order applications by name.
  NAME

  # Order applications by description.
  DESCRIPTION

  # Order applications by active.
  ACTIVE
}

# An application parameter list order.
input ApplicationParameterListOrder {

  # Field to order application parameter by.
  field: ApplicationParameterListOrderField

  # Ordering direction.
  direction: ListOrderDirection
}

# A list of fields to order an application parameter list.
enum ApplicationParameterListOrderField {

  # Order application parameters by name.
  NAME

  # Order application parameters by description.
  DESCRIPTION

  # Order application parameters by data type.
  DATATYPE

  # Order application parameters by direction.
  DIRECTION
}

`;

module.exports = applicationTypeDefs;
