/**
 * delegation.js
 * GraphQL Type definitions using schema language
 */
'use strict';

const delegationTypeDefs = `

# A delegation.
type Delegation implements Node {

  # Delegation identifier.
  id: ID!

  # User who delegates.
  delegator: User!

  # User who is designated by the delegator.
  delegatee: User
 
  # Identifies the date and time when the delegation starts.
  start: DateTime

  # Identifies the date and time when the delegation ends.
  end: DateTime

  # Process associated to the delegation.
  process: Process

  # Participant associated to the delegation.
  participant: Participant
}

# A DelegableProcessParticipant
type DelegableProcessParticipant implements Node {

  # Delegable process and participant identifier.
  id: ID!

  # Process.
  process: Process!

  # Participant.
  participant: Participant!
}

# A paginated list of delegations.
type DelegationListPage {

  # Identifies the total count of items in the page.
  totalCount: Int!

  # Indicates if there are more pages to fetch.
  hasNextPage: Boolean!

  # Indicates if there are any pages prior to the current page.
  hasPreviousPage: Boolean!

  # List of delegations.
  items: [Delegation]
}

# A paginated list of delegable process participants.
type DelegableProcessParticipantListPage {
   # Identifies the total count of items in the page.
  totalCount: Int!

  # Indicates if there are more pages to fetch.
  hasNextPage: Boolean!

  # Indicates if there are any pages prior to the current page.
  hasPreviousPage: Boolean!
  
  # List of delegable process participants.
  items: [DelegableProcessParticipant]
}

# A delegation list order.
input DelegationListOrder {

  # Field to order delegations by.
  field: DelegationListOrderField

  # Ordering direction.
  direction: ListOrderDirection
}

# A delegable process participant list order.
input DelegableProcessParticipantOrder {

  # Field to order delegable process participants by
  field: DelegableProcessParticipantField

  # Ordering direction.
  direction: ListOrderDirection
}

# A list of fields to order a delegable process participant list.
enum DelegableProcessParticipantField {

  # Order by process name.
  PROCESS_NAME

  # Order by participant name.
  PARTICIPANT_NAME
}

# A list of fields to order a delegation list.
enum DelegationListOrderField {

  # Order by delegation start date.
  START_DATE

  # Order by delegation end date.
  END_DATE
  
  # Order by delegatee name.
  DELEGATEE_NAME

  # Order by process name.
  PROCESS_NAME

  # Order by participant name.
  PARTICIPANT_NAME
}

# A filter on delegations.
input DelegationFilter {

  # Indicates if the delegations are active.
  isActive: Boolean
}

# A filter on delegations.
input DelegableProcessParticipantFilter {

  # Filter by process identifier.
  processId: ID

  # Filter by participant identifier.
  participantId: ID

  # Filter by category identifier.
  categoryId: ID
}

# A create delegation input.
input CreateDelegationInput {

  # Client mutation identifier.
  clientMutationId: String

  # Identifies the user who is the delegator.
  delegatorId: ID

  # Identifies the user who is the delegatee.
  delegateeId: ID!

  # Identifies the process associated to the delegation.
  processId: ID

  # Identifies the participant associated to the delegation.
  participantId: ID

  # Indicates if the delegatee has to be notified.
  notifyDelegatee: Boolean

  # Indicates the date and time when the delegation starts.
  start: DateTime

  # Indicates the date and time when the delegation ends.
  end: DateTime
}

# A create delegation result.
type CreateDelegationResult {

  # Client mutation identifier.
  clientMutationId: String

  # The delegation that has been created.
  delegation: Delegation!
}

# A update delegation input.
input UpdateDelegationInput {

  # Client mutation identifier.
  clientMutationId: String
  
  # Delegation identifier.
  id: ID!
  
  # Identifies the user who is the delegator.
  delegatorId: ID
  
  # Identifies the user who is the delegatee.
  delegateeId: ID
  
  # Identifies the process associated to the delegation.
  processId: ID

  # Identifies the participant associated to the delegation.
  participantId: ID

  # Indicates if the delegatee has to be notified.
  notifyDelegatee: Boolean

  # Indicates the date and time when the delegation starts.
  start: DateTime

  # Indicates the date and time when the delegation ends.
  end: DateTime
}

# A update delegation result.
type UpdateDelegationResult {
  
  # Client mutation identifier.
  clientMutationId: String
  
  # The delegation that has been updated.
  delegation: Delegation!
}

# A delete delegation input.
input DeleteDelegationInput {
  
  # Client mutation identifier.
  clientMutationId: String
  
  # Delegation identifier.
  id: ID!
}

# A delete delegation result.
type DeleteDelegationResult {

  # Client mutation identifier.
  clientMutationId: String
}

`;

module.exports = delegationTypeDefs;
