/**
 * group.js
 * GraphQL Type definitions using schema language
 */
'use strict';

const groupTypeDefs = `

# A Group.
type Group implements Node {

  # Group's identifier.
  id: ID!

  # Group's name.
  name: String!

  # Group's description.
  description: String

  # Group's query.
  query: String

  # Group's email.
  email: String

  # Group's code.
  code: String

  # Group's Ldap Active Directory's path.
  ldapADsPath: String

  # Group's common name.
  commonName: String

  # Group's distinguished name.
  distinguishedName: String

  # Group's system identifier.
  systemIdentifier: String
    
  # Is the group synchronized?
  isSynchronized: Boolean

  # Group's directory.
  directory: Directory

  # Group's list users.
  users (page: Page, orderBy: UserListOrder, filter: UserFilter): UserListPage!
}

# A paginated list of groups.
type GroupListPage {

  # Identifies the total count of items in the page.
  totalCount: Int!

  # Indicates if there are more pages to fetch.
  hasNextPage: Boolean!

  # Indicates if there are any pages prior to the current page.
  hasPreviousPage: Boolean!

  # A list of groups.
  items: [Group]
}

# A group list order.
input GroupListOrder {

  # The field to order groups by.
  field: GroupListOrderField

  # Ordering direction.
  direction: ListOrderDirection
}

# A list of fields to order a group list.
enum GroupListOrderField {

  # Order groups by name.
  NAME
}

# A create group input.
input CreateGroupInput {

  # Client mutation identifier.
  clientMutationId: String

  # Group's name.
  name: String!

  # Group's description.
  description: String

  # Group's directory identifier.
  directoryId: ID!

  # Group's query.
  query: String

  # Group's email.
  email: String

  # Group code.
  code: String

  # Group's Ldap Active Directory's path.
  ldapADsPath: String

  # Group's common name.
  commonName: String

  # Group's distinguished name.
  distinguishedName: String

  # Group's system identifier.
  systemIdentifier: String
    
  # Is the group synchronized?
  isSynchronized: Boolean
}

# A create group result.
type CreateGroupResult {
  
  # Client mutation identifier.
  clientMutationId: String

  # Group that has been created.
  group: Group!
}

# An update group input.
input UpdateGroupInput {

  # Client mutation identifier.
  clientMutationId: String

  # Group's identifier.
  id: ID!

  # Group's name.
  name: String

  # Group's description.
  description: String

  # Group's directory identifier.
  directoryId: ID

  # Group's query.
  query: String

  # Group's email.
  email: String

  # Group code.
  code: String

  # Group's Ldap Active Directory's path.
  ldapADsPath: String

  # Group's common name.
  commonName: String

  # Group's distinguished name.
  distinguishedName: String

  # Group's system identifier.
  systemIdentifier: String

  # Is the group synchronized?
  isSynchronized: Boolean
}

# An update group result.
type UpdateGroupResult {

  # Client mutation identifier.
  clientMutationId: String

  # The group that has been updated.
  group: Group!
}

# A delete group input.
input DeleteGroupInput {

  # Client mutation identifier.
  clientMutationId: String  

  # Group's identifier.
  id: ID!
}

# A delete group result.
type DeleteGroupResult {
  
  # Client mutation identifier.
  clientMutationId: String    
}

# An add users to group input.
input AddUsersToGroupInput {

  # Client mutation identifier.
  clientMutationId: String

  # Group's identifier.
  id: ID!

  # Users' ids to be added to the group
  userIds: [ID]!
}

# An add users to group result.
type AddUsersToGroupResult {

  # Client mutation identifier.
  clientMutationId: String

  # The group that has been updated.
  group: Group!
}

# A remove users from group input.
input RemoveUsersFromGroupInput {

  # Client mutation identifier.
  clientMutationId: String  

  # Group's identifier.
  id: ID!

  # Users' ids to be removed from the group.
  userIds: [ID]!
}

# A remove users from group result.
type RemoveUsersFromGroupResult {
  
  # Client mutation identifier.
  clientMutationId: String
    
  # The group that has been updated.
  group: Group!
}

`;

module.exports = groupTypeDefs;
