/**
 * request.js
 * GraphQL Type definitions using schema language
 */
'use strict';

const requestTypeDefs = `

# A Request.
type Request implements Node {

  # Request identifier.
  id: ID!

  # Request name.
  name: String!

  # Request description.
  description: String!

  # Request number.
  number: Int!

  # Request status.
  status: RequestStatus!

  # Request sub-status.
  subStatus: RequestSubStatus!

  # Indicates if the request is in test.
  isTest: Boolean!

  # Parent action of the request.
  parentAction: Action

  # Process associated to the request.
  process: Process!

  # Actions associated to the request.
  actions(page: Page, orderBy: ActionListOrder, filter: RequestActionFilter): ActionListPage!
 
  # Identifies the date limit of the request.
  limit: DateTime

  # Identifies the user who initiated the request.
  requester: User!

  # Identifies the delegated user who initiated the request.
  delegatedRequester: User

  # Identifies the date and time when the request was opened.
  openedAt: DateTime!

  # Identifies the date and time when the request was started.
  startedAt: DateTime

  # Identifies the date and time when the request was started.
  closedAt: DateTime

  # Identifies the user who aborted the request.
  abortedBy: User

  # URL of the request follow-up form.
  followUpFormUrl: URI!

  # URL of the request follow-up form displayed in the portal.
  followUpFormPortalUrl: URI!

  # URL of the request graphical follow-up.
  graphicalFollowUpUrl: URI!

  # URL of the request graphical follow-up displayed in the portal.
  graphicalFollowUpPortalUrl: URI!

  # Process data associated to the request.
  dataset (page: Page, orderBy: RequestDataListOrder, filter: RequestDataFilter) : RequestDataListPage!
  
  # Comments associated to the request.
  comments (page: Page, orderBy: CommentListOrder) : CommentListPage!

  # Indicates if the user can cancel the request.
  canCancel: Boolean
  
  # Indicates if the user can delete the request.
  canDelete: Boolean
}

# Represent the possible request statuses.
enum RequestStatus {

  # Request that is still open.
  OPEN

  # Request that has been closed.
  CLOSED
}

# Represent the possible request sub-statuses.
enum RequestSubStatus {

  # Request that is still not running and not started.
  NOT_RUNNING_NOT_STARTED

  # Request that is still running.
  RUNNING

  # Request that has been completed.
  COMPLETED

  # Request that has been aborted.
  ABORTED 
}

# A paginated list of requests.
type RequestListPage {

  # Identifies the total count of items in the page.
  totalCount: Int!

  # Indicates if there are more pages to fetch.
  hasNextPage: Boolean!

  # Indicates if there are any pages prior to the current page.
  hasPreviousPage: Boolean!

  # List of requests.
  items: [Request]
}

# A filter on requests.
input RequestFilter {

  # Filter requests by process name.
  processName: String
  
  # Filter requests by process version.
  processVersion: Int

  # Filter requests by status.
  status: RequestStatus
  
  # Filter requests by a list of sub-status.
  subStatuses: [RequestSubStatus]

  # Indicates if requests are in test.
  isTest: Boolean

  # Indicates if requests are or were late.
  isDueDateExceeded: Boolean

  # Filter requests by role.
  as: RequestRole

  # Filter requests by requester identifier.
  requesterId: ID

  # Filter requests by participant identifier.
  participantId: ID
  
  # Filter requests by creation date since.
  createdSince: DateTime
}

# A request list order.
input RequestListOrder {

  # Field to order requests by.
  field: RequestListOrderField

  # Ordering direction.
  direction: ListOrderDirection
}

# A list of fields to order a list of requests.
enum RequestListOrderField {

  # Order the requests by number.
  NUMBER

  # Order the requests by when the request was opened.
  OPENED_AT

  # Order the requests by when the request was closed.
  CLOSED_AT

  # Order the requests by date limit.
  LIMIT

  # Order the requests by status.
  STATUS
}

# Represent the possible request roles.
enum RequestRole {

  # An user who has initiated a request.
  REQUESTER

  # An user who is a supervisor of a request.
  SUPERVISOR

}

# A create request input.
input CreateRequestInput {
  
  # Client mutation identifier.
  clientMutationId: String
  
  # Process identifier. 
  processId: ID
  
  # Process name. 
  processName: String
  
  # Process version.
  processVersion: Int
  
  # Indicates if the process is in test.
  isTest: Boolean
  
  # User identifier that is used by the user to perform the mutation on behalf of another person.
  onBehalfOf: ID

  # Array of parameters.
  parameters: [ParameterInput]
}

# A create request result.
type CreateRequestResult {
  
  # Client mutation identifier.
  clientMutationId: String
  
  # Request that has been created.
  request: Request!
}

# A cancel request input.
input CancelRequestInput {
  
  # Client mutation identifier.
  clientMutationId: String
  
  # Request identifier.
  id: ID
  
  # Request number.
  number: Int

  # User identifier that is used by the user to perform the mutation on behalf of another person.
  onBehalfOf: ID
  
  # Indicates if the participant members have to be notified.
  notifyParticipant: Boolean  
}

# A cancel request result.
type CancelRequestResult {
  
  # Client mutation identifier.
  clientMutationId: String
  
  # Request that has been cancelled.
  request: Request!
}

# A delete request input.
input DeleteRequestInput {
  
  # Client mutation identifier.
  clientMutationId: String

  # Request identifier.
  id: ID

  # Request number.
  number: Int
}

# A delete request result.
type DeleteRequestResult {
  
  # Client mutation identifier.
  clientMutationId: String
}

`;

module.exports = requestTypeDefs;
