/**
 * user.js
 * GraphQL Type definitions using schema language
 */
'use strict';

const userTypeDefs = `

# A user.
type User implements Node {

  # User's identifier.
  id: ID!

  # User's login name.
  userName: String!

  # User's last name.
  lastName: String

  # User's first name.
  firstName: String

  # User's email.
  email: String

  # User's phone number.
  phone: String

  # User's mobile phone number.
  mobile: String

  # User's pager number.
  pager: String

  # User's fax number.
  fax: String

  # User's company.
  company: String

  # User's department.
  department: String

  # User's office.
  office: String

  # User's job title.
  jobTitle: String

  # User's employee number.
  employeeNumber: String

  # User's employee type.
  employeeType: String

  # User's manager.
  manager: User

  # User's personal title.
  personalTitle: String

  # User's initials.
  initials: String

  # User's postal address.
  postalAddress: String

  # User's postal code.
  postalCode: String

  # User's city.
  city: String

  # User's state.
  state: String

  # User's country.
  country: String

  # User's default language.
  defaultLanguage: String!

  # User's preferred language.
  language: String!

  # User's Ldap Active Directory's path.
  ldapADsPath: String

  # User's common name.
  commonName: String

  # User's distinguisehd name.
  distinguishedName: String

  # User's system identifier.
  systemIdentifier: String

  # User's extended attribute 1.
  extendedAttribute1: String
  
  # User's extended attribute 2.
  extendedAttribute2: String
  
  # User's extended attribute 3.
  extendedAttribute3: String

  # User's extended attribute 4.
  extendedAttribute4: String

  # User's extended attribute 5.
  extendedAttribute5: String

  # Is the user a process supervisor?
  isSupervisor: Boolean!

  # Is the user a process folder manager?
  isProcessFolderManager: Boolean!

  # Is the user an administrator?
  isAdministrator: Boolean!

  # Is the user's account active?
  isActive: Boolean!

  # Is the user's account archived?
  isArchived: Boolean!

  # Is the user the viewing user?
  isViewer: Boolean!
    
  # Is the user's account synchronized?
  isSynchronized: Boolean
  
  # User's default time zone identifier.
  defaultTimeZoneId: Int!

  # User's preferred time zone identifier.
  timeZoneId: Int!

  # User's directory.
  directory: Directory

  # List of the user's groups.
  groups (page: Page, orderBy: GroupListOrder): GroupListPage!

  # List of processes as participant.
  processes (page: Page,  orderBy: ProcessListOrder, filter: ProcessFilter, onBehalfOf: ID): ProcessListPage!

  # List of processes that the user can launch. 
  launchableProcesses (page: Page, orderBy: ProcessListOrder, filter: LaunchableProcessFilter, onBehalfOf: ID): ProcessListPage!

  # List of requests as participant.
  requests (page: Page, orderBy: RequestListOrder, filter: RequestFilter, onBehalfOf: ID, viewerAsMember: Boolean): RequestListPage!

  # List of actions as participant.
  actions (page: Page, orderBy: ActionListOrder, filter: ActionFilter, onBehalfOf: ID, viewerAsMember: Boolean): ActionListPage!

  # List of comments as participant.
  comments (page: Page, orderBy: CommentListOrder, onBehalfOf: ID, viewerAsMember: Boolean): CommentListPage!

  # List of participants as member.
  participants (page: Page, orderBy: ParticipantListOrder, filter: ParticipantFilter, onBehalfOf: ID): ParticipantListPage!

  # List of user's favorites
  favorites (page: Page, orderBy: FavoriteListOrder, filter: FavoriteFilter) : FavoriteListPage!

  # List of views available to the user.
  views (page: Page, orderBy: ViewListOrder, filter: ViewFilter): ViewListPage!

  # List of delegations managed by the user.
  delegations (page: Page, orderBy: DelegationListOrder, filter: DelegationFilter): DelegationListPage!

  # List of the user's delegators.
  delegators (page: Page): UserListPage!

  # List of the user's delegable process participants.
  delegableProcessParticipants (page: Page, orderBy: DelegableProcessParticipantOrder, filter: DelegableProcessParticipantFilter): DelegableProcessParticipantListPage!
}

# A paginated list of users.
type UserListPage {

  # Identifies the total count of items in the page.
  totalCount: Int!

  # Indicates if there are more pages to fetch.
  hasNextPage: Boolean!

  # Indicates if there are any pages prior to the current page.
  hasPreviousPage: Boolean!

  # A list of users.
  items: [User]
}

# A user list order.
input UserListOrder {

  # The field to order users by.
  field: UserListOrderField

  # Ordering direction.
  direction: ListOrderDirection
}

# A list of fields to order a user list.
enum UserListOrderField {

  # Order users by username.
  USERNAME

  # Order users by last name.
  LASTNAME
}

# A create user input.
input CreateUserInput {
  
  # Client mutation identifier.
  clientMutationId: String

  # User's login name.
  userName: String!

  # User's last name.
  lastName: String!

  # User's first name.
  firstName: String

  # Directory identifier.
  directoryId: ID!

  # Is the user an administrator?
  isAdministrator: Boolean
  
  # User's password.
  password: String

  # Is the user's account active?
  isActive: Boolean

  # Is the user's account synchronized?
  isSynchronized: Boolean

  # User's email.
  email: String
    
  # User's default language.
  defaultLanguage: String

  # User's default time zone identifier.
  defaultTimeZoneId: Int

  # User's manager identifier.
  managerId: ID

  # User's phone number.
  phone: String
  
  # User's mobile phone number.
  mobile: String

  # User's fax number.
  fax: String

  # User's pager number.
  pager: String
    
  # User's office.
  office: String

  # User's department.
  department: String
  
  # User's company.
  company: String

  # User's job title.
  jobTitle: String

  # User's employee number.
  employeeNumber: String

  # User's employee type.
  employeeType: String

  # User's initials.
  initials: String

  # User's personal title.
  personalTitle: String

  # User's postal address.
  postalAddress: String

  # User's postal code.
  postalCode: String

  # User's city.
  city: String

  # User's state.
  state: String

  # User's country.
  country: String

  # User's Ldap Active Directory's path.
  ldapADsPath: String

  # User's common name.
  commonName: String

  # User's distinguisehd name.
  distinguishedName: String

  # User's extended attribute 1.
  extendedAttribute1: String
  
  # User's extended attribute 2.
  extendedAttribute2: String
  
  # User's extended attribute 3.
  extendedAttribute3: String

  # User's extended attribute 4.
  extendedAttribute4: String

  # User's extended attribute 5.
  extendedAttribute5: String

  # User's system identifier.
  systemIdentifier: String 
}

# A create user result.
type CreateUserResult {
  
  # Client mutation identifier.
  clientMutationId: String
  
  # User that has been created.
  user: User!
}

# An update user input.
input UpdateUserInput {
    
  # Client mutation identifier.
  clientMutationId: String

  # User's identifier.
  id: ID!

  # User's login name.
  userName: String

  # User's last name.
  lastName: String

  # User's first name.
  firstName: String

  # Directory identifier.
  directoryId: ID

  # Is the user an administrator?
  isAdministrator: Boolean
  
  # User's password.
  password: String

  # Is the user's account active?
  isActive: Boolean

  # Is the user's account synchronized?
  isSynchronized: Boolean

  # User's email.
  email: String
    
  # User's default language.
  defaultLanguage: String

  # User's preferred language.
  language: String

  # User's default time zone identifier.
  defaultTimeZoneId: Int

  # User's preferred time zone identifier.
  timeZoneId: Int

  # User's manager identifier.
  managerId: ID

  # User's phone number.
  phone: String
  
  # User's mobile phone number.
  mobile: String

  # User's fax number.
  fax: String

  # User's pager number.
  pager: String
    
  # User's office.
  office: String

  # User's department.
  department: String
  
  # User's company.
  company: String

  # User's job title.
  jobTitle: String

  # User's employee number.
  employeeNumber: String

  # User's employee type.
  employeeType: String

  # User's initials.
  initials: String

  # User's personal title.
  personalTitle: String

  # User's postal address.
  postalAddress: String

  # User's postal code.
  postalCode: String

  # User's city.
  city: String

  # User's state.
  state: String

  # User's country.
  country: String

  # User's Ldap Active Directory's path.
  ldapADsPath: String

  # User's common name.
  commonName: String

  # User's distinguisehd name.
  distinguishedName: String

  # User's extended attribute 1.
  extendedAttribute1: String
  
  # User's extended attribute 2.
  extendedAttribute2: String
  
  # User's extended attribute 3.
  extendedAttribute3: String

  # User's extended attribute 4.
  extendedAttribute4: String

  # User's extended attribute 5.
  extendedAttribute5: String

  # User's system identifier.
  systemIdentifier: String
}

# An update user result.
type UpdateUserResult {
  
  # Client mutation identifier.
  clientMutationId: String
  
  # The user that has been updated.
  user: User!
}

# A delete user input.
input DeleteUserInput {
  
  # Client mutation identifier.
  clientMutationId: String
  
  # User's identifier.
  id: ID!
}

# A delete user result.
type DeleteUserResult {

  # Client mutation identifier.
  clientMutationId: String
}

# An add groups to user input.
input AddGroupsToUserInput {

  # Client mutation identifier.
  clientMutationId: String

  # User's identifier.
  id: ID!

  # Group' ids to be added to the user.
  groupIds: [ID]!
}

# An add groups to user result.
type AddGroupsToUserResult {

  # Client mutation identifier.
  clientMutationId: String

  # The user that has been updated.
  user: User!
}

# A remove groups from user input.
input RemoveGroupsFromUserInput {

  # Client mutation identifier.
  clientMutationId: String  

  # User's identifier.
  id: ID!

  # Groups' ids to be removed from the user.
  groupIds: [ID]!
}

# A remove groups from user result.
type RemoveGroupsFromUserResult {
  
  # Client mutation identifier.
  clientMutationId: String
    
  # The user that has been updated.
  user: User!
}

# A filter on users.
input UserFilter {

  # Filter by user's first name and last name.
  nameContains: String

  # Filter by active users.
  isActive: Boolean

  # Filter by archived users.
  isArchived: Boolean

}

# An update viewer preferences input.
input UpdateViewerPreferencesInput {

  # Client mutation identifier.
  clientMutationId: String
  
  # Viewer's preferred language.
  language: String

  # Viewer's preferred time zone identifier.
  timeZoneId: Int
}

# An update viewer preferences result.
type UpdateViewerPreferencesResult {
  
  # Client mutation identifier.
  clientMutationId: String
  
  # The user that has been updated.
  user: User!
}

`;

module.exports = userTypeDefs;
