/**
 * server.js
 * Server initialization
 */
'use strict';
const express = require('express');
const cors = require('cors');
const graphqlExpress = require('express-graphql');
const schemaPrinter = require('graphql/utilities/schemaPrinter');
const { createSchema } = require('./models/schema');
const uuidV1 = require('uuid/v1');
const WorkflowService = require('./services/workflow/workflowservice');
const { logger } = require('./utils/logger.js');
const { authenticate, getWorkflowContext } = require('./models/auth/context');
const { createLoaders } = require('./models/loaders/');
const config = require('./config');
const bodyParser = require('body-parser');

const rootUrl = '(\/\\w+)?\/graphql'; // eslint-disable-line
const app = express();
var router = express.Router();

logger('INFO', 'Application NODE_ENV: ' + process.env.NODE_ENV);
logger('INFO', 'GraphQL debug enabled: ' + config.graphql.debugEnabled);
logger('INFO', 'Workflow service debug enabled: ' + config.workflowService.debugEnabled);

// Schema creation
const schema = createSchema(config.graphql);

// Workflow service creation
const workflowService = new WorkflowService(config.workflowService);

// Enable Cross Origin Resource Sharing
app.use('*', cors());

// Parse application/x-www-form-urlencoded
app.use(bodyParser.urlencoded({ limit: `${config.workflowService.maxInputFileContentSize + config.graphql.maxRequestBodySize}kb`, extended: false }));

// Parse application/json
app.use(bodyParser.json({ limit: `${config.workflowService.maxInputFileContentSize + config.graphql.maxRequestBodySize}kb` }));

// Catch body content size error
app.use((err, req, res, next) => {
  if (err.type === 'entity.too.large') {
    logger('ERROR', err.message);
    return res.status(400).json({ error: err.message });
  } else {
    next();
  }
});

// Authorization: an authentication is required
app.use((req, res, next) => {
  req.uuid = uuidV1();
  authenticate(req)
    .then(sub => {
      logger('DEBUG', 'Request ' + req.uuid + ' Authenticate: sub: ' + sub);
      next();
    })
    .catch(error => {
      logger('ERROR', 'Request ' + req.uuid + ' Unauthorized: unable to get the authenticated username. Error: ' + error);
      res.status(401);
      res.send('Unauthorized: Access is denied due to missing or invalid credentials.');
    });
});

// Print Schema endpoint
router.use('/schema', (req, res) => {
  res.set('Content-Type', 'text/plain');
  res.send(schemaPrinter.printSchema(schema));
});

// Print Introspection Schema endpoint
router.use('/ischema', (req, res) => {
  res.set('Content-Type', 'text/plain');
  res.send(schemaPrinter.printIntrospectionSchema(schema));
});

// GraphQL endpoint
router.use(graphqlExpress(request => {
  const startTime = Date.now();

  request.workflow = {
    service: workflowService,
    context: getWorkflowContext(request)
  };
  request.loaders = createLoaders(config.loaders, request);
  request.resolverCount = 0;
  request.logTimeoutError = true;

  logger('INFO', 'new request ' + request.uuid + ' by ' + request.workflow.context.authUserName);

  request.incrementResolverCount = function () {
    var runTime = Date.now() - startTime;
    if (runTime > config.graphql.queryTimeout) {
      if (request.logTimeoutError) {
        logger('ERROR', 'Request ' + request.uuid + ' query execution timeout');
      }
      request.logTimeoutError = false;
      throw new Error('Query execution has timeout. Field resolution aborted.');
    }
    this.resolverCount++;
  };

  return !config.graphql.debugEnabled
    ? {
      schema: schema,
      context: request,
      graphiql: config.graphql.graphiqlEnabled
    }
    : {
      schema: schema,
      context: request,
      graphiql: config.graphql.graphiqlEnabled,
      formatError: error => ({
        message: error.message,
        locations: error.locations,
        stack: error.stack
      }),
      extensions ({ document, variables, operationName, result }) {
        return {
          requestId: request.uuid,
          runTime: Date.now() - startTime,
          resolverCount: request.resolverCount,
          operationCount: request.workflow.context.operationCount,
          operationErrorCount: request.workflow.context.operationErrorCount
        };
      }
    };
}));

app.use(rootUrl, router);

// Start Express Server
app.listen(config.app.port, function () {
  logger('INFO', 'GraphQL Server running on port ' + config.app.port);
});
