/**
 * groups.js
 * Controller for the /Groups endpoint.
 */
'use strict';

const constants = require('../models/constants');
const express = require('express');
const router = express.Router(); // eslint-disable-line
const ErrorHelper = require('../models/helpers/error');
const Directory = require('../models/directory');
const SearchOperation = require('../models/search');
const PatchOperations = require('../models/patchoperations.js');
const { logToFile, Actions } = require('../models/helpers/logger');

/**
 * GET /Groups
 * @desc Performs a search through a GET request.
 */
router.get('/', (req, res) => {
  const directory = new Directory(null, req.authToken);
  const attributes = req.query.attributes;
  const excludedAttributes = req.query.excludedAttributes;

  logToFile(Actions.GROUP, directory, 'Search groups with filter: %s', req.query.filter);
  directory.search(
    constants.schemas.group,
    req.query.filter,
    req.query.sortBy,
    req.query.sortOrder ? req.query.sortOrder.toLocaleLowerCase() : null,
    req.query.startIndex,
    req.query.count,
    !attributes || attributes instanceof Array
      ? attributes : attributes.split(','),
    !excludedAttributes || excludedAttributes instanceof Array
      ? excludedAttributes : excludedAttributes.split(',')
  )
    .then(results => {
      logToFile(Actions.GROUP, directory, 'Search groups: completed (filter: %s)', req.query.filter);
      res.status(200).json({
        schemas: [constants.schemas.listResponse],
        Resources: results,
        totalResults: results.length,
        itemsPerPage: req.query.count > 0 ? req.query.count : results.length,
        startIndex: req.query.startIndex
      });
    })
    .catch(err => {
      if (err instanceof SearchOperation.InvalidFilterError) {
        logToFile(Actions.GROUP_ERROR, directory, 'Invalid filter %s', req.query.filter);
        return ErrorHelper.sendInvalidFilterError(res);
      }

      if (err instanceof SearchOperation.InvalidSortOrderError) {
        logToFile(Actions.GROUP_ERROR, directory, 'Invalid sort order provided (%s)', req.query.sortOrder);
        return ErrorHelper.sendInvalidValue(res, err.message);
      }

      if (err instanceof SearchOperation.InvalidCountError) {
        logToFile(Actions.GROUP_ERROR, directory, 'Invalid count provided (%s)', req.query.count);
        return ErrorHelper.sendInvalidValue(res, err.message);
      }

      if (err.statusCode) {
        return ErrorHelper.handleWorkflowServiceErrors(err, res, Actions.GROUP_ERROR, directory);
      }

      logToFile(Actions.GROUP_ERROR, directory, 'Unexpected error: %s (filter: %s)', err.message, req.params.filter);
      return ErrorHelper.sendInternalServerError(res, err.message);
    });
});

/**
 * GET /Groups/:groupId
 * @desc Gets the group associated with {groupId}.
 * @param {String} groupId The id of the group.
 */
router.get('/:groupId', (req, res) => {
  const groupId = req.params.groupId;
  const attributes = 'attributes' in req.query
    ? req.query.attributes.split(',') : null;
  const excludedAttributes = 'excludedAttributes' in req.query
    ? req.query.excludedAttributes.split(',') : null;
  const baseUrl = req.getBaseUrl();
  const directory = new Directory(baseUrl, req.authToken);

  logToFile(Actions.GROUP, directory, 'GET Group: %s', groupId);
  directory
    .getGroup(groupId, attributes, excludedAttributes)
    .then(repr => {
      logToFile(Actions.GROUP, directory, 'GET Group: completed (groupId: %s)', groupId);
      res.location((repr.meta && repr.meta.location) || baseUrl);
      res.status(200).json(repr);
    })
    .catch(err => {
      if (err.statusCode) {
        return ErrorHelper.handleWorkflowServiceErrors(err, res, Actions.GROUP_ERROR, directory);
      }

      logToFile(Actions.GROUP_ERROR, directory, `Unexpected error: ${(err && err.message) || 'no message'} (groupId: %s)`, groupId);
      ErrorHelper.sendInternalServerError(res);
    });
});

/**
 * DELETE /Groups/:groupId
 * @desc Deletes a group associated with {groupId}
 * @param {String} groupId The id of the group.
 */
router.delete('/:groupId', (req, res) => {
  const groupId = req.params.groupId;
  const baseUrl = req.getBaseUrl();
  const directory = new Directory(baseUrl, req.authToken);

  logToFile(Actions.GROUP_DELETE, directory, 'DELETE Group (groupId: %s)', groupId);
  directory
    .deleteGroup(groupId)
    .then(() => {
      logToFile(Actions.GROUP_DELETE, directory, 'DELETE Group: completed (groupId: %s)', groupId);
      res.sendStatus(204);
    })
    .catch(err => {
      if (err.statusCode) {
        return ErrorHelper.handleWorkflowServiceErrors(err, res, Actions.GROUP_ERROR, directory);
      }

      logToFile(Actions.GROUP_ERROR, directory, 'Unexpected error: %s (groupId: %s)', err.message, groupId);
      ErrorHelper.sendInternalServerError(res, err.message);
    });
});

/**
 * POST /Groups/
 * @desc Creates a group.
 */
router.post('/', (req, res) => {
  const baseUrl = req.getBaseUrl();
  const directory = new Directory(baseUrl, req.authToken);

  logToFile(Actions.GROUP_ADD, directory, 'POST Group (displayName: %s)', req.body.displayName);
  if (
    !req.body.schemas ||
    !req.body.schemas.includes(constants.schemas.group) ||
    !req.body.displayName ||
    req.body.displayName === ''
  ) {
    logToFile(Actions.GROUP_ERROR, directory, 'Invalid syntax (displayName: %s)', req.body.displayName);
    ErrorHelper.sendInvalidSyntax(res);
    return;
  }

  directory
    .createGroup(req.body)
    .then(repr => {
      logToFile(Actions.GROUP_ADD, directory, 'POST Group: completed (displayName: %s)', req.body.displayName);
      res.location(repr.meta.location);
      res.status(201).json(repr);
    })
    .catch(err => {
      if (err.statusCode) {
        ErrorHelper.handleWorkflowServiceErrors(err, res, Actions.GROUP_ERROR, directory);
      } else {
        logToFile(Actions.GROUP_ERROR, directory, 'Unexpected error (displayName: %s)', req.body.displayName);
        ErrorHelper.sendInternalServerError(res, err);
      }
    });
});

/**
 * PATCH /Groups/:groupId
 * @desc Partially modifies a group associated with {groupId}.
 * @param {String} groupId The id of the group.
 */
router.patch('/:groupId', (req, res) => {
  const baseUrl = req.getBaseUrl();
  const directory = new Directory(baseUrl, req.authToken);

  logToFile(Actions.GROUP_UPDATE, directory, 'PATCH Group (groupId: %s)', req.params.groupId);

  if (
    !req.body.schemas ||
    !req.body.schemas.includes(constants.schemas.patchOperation)
  ) {
    logToFile(Actions.GROUP_ERROR, directory, 'Incorrect schema (groupId: %s)', req.params.groupId);
    return ErrorHelper.sendIncorrectSchema(res);
  } else if (
    !req.body.Operations ||
    !(req.body.Operations instanceof Array) ||
    req.body.Operations.length <= 0
  ) {
    logToFile(Actions.GROUP_ERROR, directory, 'Missing patch operations (groupId: %s)', req.params.groupId);
    return ErrorHelper.sendPatchOperationsMissing(res);
  }

  directory
    .partialUpdateGroup(req.params.groupId, req.body.Operations)
    .then(repr => {
      logToFile(Actions.GROUP_UPDATE, directory, 'PATCH Group: completed (groupId: %s)', req.params.groupId);
      res.location(repr.meta.location);
      res.status(200).json(repr);
    })
    .catch(err => {
      if (err instanceof PatchOperations.PatchImmutableError) {
        logToFile(Actions.GROUP_ERROR, directory, 'Tried to modify an immutable value (groupId: %s)', req.params.groupId);
        ErrorHelper.sendImmutablePatchError(res);
      } else if (err instanceof PatchOperations.PatchNoPathError) {
        logToFile(Actions.GROUP_ERROR, directory, 'Missing path (groupId: %s)', req.params.groupId);
        ErrorHelper.sendNoPath(res);
      } else if (err instanceof PatchOperations.PatchNoValueError) {
        logToFile(Actions.GROUP_ERROR, directory, 'No value provided (groupId: %s)', req.params.groupId);
        ErrorHelper.sendNoValue(res);
      } else if (err instanceof PatchOperations.PatchOpMissingError) {
        logToFile(Actions.GROUP_ERROR, directory, 'No operations provided (groupId: %s)', req.params.groupId);
        ErrorHelper.sendPatchOperationsMissing(res);
      } else if (err instanceof PatchOperations.PatchInvalidFilterError) {
        logToFile(Actions.GROUP_ERROR, directory, 'Invalid filter (groupId: %s)', req.params.groupId);
        ErrorHelper.sendInvalidFilterError(res);
      } else if (err instanceof PatchOperations.PatchNoTargetError) {
        logToFile(Actions.GROUP_ERROR, directory, 'Not target provided (groupId: %s)', req.params.groupId);
        ErrorHelper.sendNoPathReplace(res);
      } else if (err instanceof PatchOperations.PatchRemoveComplexAttributeError) {
        logToFile(Actions.GROUP_ERROR, directory, 'Tried to remove a complexe attribute (groupId: %s)', req.params.groupId);
        ErrorHelper.sendPatchRemoveComplexAttributeError(res);
      } else if (err.statusCode) {
        ErrorHelper.handleWorkflowServiceErrors(err, res, Actions.GROUP_ERROR, directory);
      } else {
        logToFile(Actions.GROUP_ERROR, directory, 'Unexpected error: %s (groupId: %s)', err.message, req.params.groupId);
        ErrorHelper.sendInternalServerError(res);
      }
    });
});

module.exports = router;
