/**
 * User.js
 * Controller for the /user endpoint.
 */
'use strict';
const express = require('express');
const router = express.Router();  // eslint-disable-line
const ErrorHelper = require('../models/helpers/error');
const constants = require('../models/constants');
const PatchOperations = require('../models/patchoperations.js');
const SearchOperation = require('../models/search');
const Directory = require('../models/directory');
const { logToFile, Actions } = require('../models/helpers/logger');

/**
 * GET /Users
 * @desc Performs a search through a GET request.
 */
router.get('/', (req, res) => {
  const directory = new Directory(null, req.authToken);
  const attributes = req.query.attributes;
  const excludedAttributes = req.query.excludedAttributes;

  logToFile(Actions.USER, directory, 'Search operation with filter: %s', req.query.filter);
  directory.search(
    constants.schemas.user,
    req.query.filter,
    req.query.sortBy,
    req.query.sortOrder ? req.query.sortOrder.toLocaleLowerCase() : null,
    req.query.startIndex,
    req.query.count,
    !attributes || attributes instanceof Array
      ? attributes : attributes.split(','),
    !excludedAttributes || excludedAttributes instanceof Array
      ? excludedAttributes : excludedAttributes.split(',')  // eslint-disable-line
  )
    .then(results => {
      logToFile(Actions.USER, directory, 'Search operation with filter (%s): completed', req.query.filter);
      res.status(200).json({
        schemas: [constants.schemas.listResponse],
        Resources: results,
        totalResults: results.length,
        itemsPerPage: req.query.count > 0 ? req.query.count : results.length,
        startIndex: req.query.startIndex
      });
    })
    .catch(err => {
      if (err instanceof SearchOperation.InvalidFilterError) {
        logToFile(Actions.USER_ERROR, directory, 'Invalid filter: %s', req.query.filter);
        return ErrorHelper.sendInvalidFilterError(res);
      }

      if (err instanceof SearchOperation.InvalidSortOrderError) {
        logToFile(Actions.USER_ERROR, directory, 'Invalid sort order: %s', req.query.sortOrder);
        return ErrorHelper.sendInvalidValue(res, err.message);
      }

      if (err instanceof SearchOperation.InvalidCountError) {
        logToFile(Actions.USER_ERROR, directory, 'Invalid count: %s', req.query.count);
        return ErrorHelper.sendInvalidValue(res, err.message);
      }

      if (err instanceof SearchOperation.TooManyResultsError) {
        logToFile(Actions.USER_ERROR, directory, 'Too many results');
        return ErrorHelper.sendTooManyQueryResults(res);
      }

      if (err.statusCode) {
        return ErrorHelper.handleWorkflowServiceErrors(err, res, Actions.USER_ERROR, directory);
      }

      logToFile(Actions.USER_ERROR, directory, 'Unexpected error: %s', err.message);
      return ErrorHelper.sendInternalServerError(res, err.message);
    });
});

/**
 * GET /Users/:userId
 * @desc Get the user associated with {userId}.
 * @param {String} userId The id of the user.
 */
router.get('/:userId', (req, res) => {
  const userId = req.params.userId;
  const attributes = 'attributes' in req.query
    ? req.query.attributes.split(',') : null;
  const excludedAttributes = 'excludedAttributes' in req.query
    ? req.query.excludedAttributes.split(',') : null;
  const baseUrl = req.getBaseUrl();
  const directory = new Directory(baseUrl, req.authToken);

  logToFile(Actions.USER, directory, 'GET user with id %s', userId);

  if (typeof userId === 'undefined' || userId === '') {
    logToFile(Actions.USER_ERROR, directory, 'User id missing');
    return ErrorHelper.sendIdMissing(res);
  }

  directory
    .getUser(userId, attributes, excludedAttributes)
    .then(repr => {
      logToFile(Actions.USER, directory, 'GET user with id %s: completed', userId);
      res.location(
        repr.meta && repr.meta.location ? repr.meta.location : baseUrl
      );
      res.status(200).json(repr);
    })
    .catch(err => {
      if (err.statusCode) {
        return ErrorHelper.handleWorkflowServiceErrors(err, res, Actions.USER_ERROR, directory);
      }

      logToFile(Actions.USER_ERROR, directory, 'Unexpected error: %s (User %s)', err.message, userId);
      ErrorHelper.sendInternalServerError(res);
    });
});

/**
 * POST /Users/
 * @desc Creates a user.
 */
router.post('/', (req, res) => {
  const hasFamilyName = typeof req.body.name === 'object'
    ? req.body.name.familyName : req.body.name;
  const baseUrl = req.getBaseUrl();
  const directory = new Directory(baseUrl, req.authToken);

  logToFile(Actions.USER_ADD, directory, 'POST user (userName: %s)', req.body.userName);

  if (
    !req.body.schemas ||
    !req.body.schemas.includes(constants.schemas.user) ||
    !req.body.userName ||
    req.body.userName === '' ||
    !hasFamilyName
  ) {
    logToFile(Actions.USER_ERROR, directory, 'Invalid syntax, (userName: %s)', req.body.userName);
    return ErrorHelper.sendInvalidSyntax(res);
  }

  directory
    .createUser(req.body)
    .then(repr => {
      logToFile(Actions.USER_ADD, directory, 'POST user: completed (userName: %s)', req.body.userName);
      res.location(repr.meta.location);
      res.status(201).json(repr);
    })
    .catch(err => {
      if (err === constants.scimTypes.uniqueness) {
        logToFile(Actions.USER_ERROR, directory, 'User already exists for userName: %s', req.body.userName);
        ErrorHelper.sendUniquenessError(res);
      } else if (err.statusCode) {
        ErrorHelper.handleWorkflowServiceErrors(err, res, Actions.USER_ERROR, directory);
      } else {
        logToFile(Actions.USER_ERROR, directory, 'Unexpected error: %s (userName: %s)', err.message, req.body.userName);
        ErrorHelper.sendInternalServerError(res, err);
      }
    });
});

/**
 * PATCH /Users/:userId
 * @desc Modifies part of the user associated with {userId}.
 * @param {userId} the id of the user.
 */
router.patch('/:userId', (req, res) => {
  const baseUrl = req.getBaseUrl();
  const directory = new Directory(baseUrl, req.authToken);

  logToFile(Actions.USER_UPDATE, directory, 'PATCH user: %s', req.params.userId);

  if (
    !req.body.schemas ||
    !req.body.schemas.includes(constants.schemas.patchOperation)
  ) {
    logToFile(Actions.USER_ERROR, directory, 'Incorrect schema (userId: %s)', req.params.userId);
    return ErrorHelper.sendIncorrectSchema(res);
  } else if (
    !req.body.Operations ||
    !(req.body.Operations instanceof Array) ||
    req.body.Operations.length <= 0
  ) {
    logToFile(Actions.USER_ERROR, directory, 'Missing operations to execute');
    return ErrorHelper.sendPatchOperationsMissing(res);
  }

  directory
    .partialUpdateUser(req.params.userId, req.body.Operations)
    .then(repr => {
      logToFile(Actions.USER_UPDATE, directory, 'PATCH user: completed (userId: %s)', req.params.userId);
      res.location(repr.meta.location);
      res.status(200).json(repr);
    })
    .catch(err => {
      if (
        err === 'uniqueness' ||
        (
          err.statusCode === 500 &&
          err.message.indexOf('already in use') !== -1
        )
      ) {
        logToFile(Actions.USER_ERROR, directory, 'Value already in use (user %s)', req.params.userId);
        ErrorHelper.sendUniquenessError(res);
      } else if (err instanceof PatchOperations.PatchImmutableError) {
        logToFile(Actions.USER_ERROR, directory, 'Value is immutable (user %s)', req.params.userId);
        ErrorHelper.sendImmutablePatchError(res);
      } else if (err instanceof PatchOperations.PatchNoPathError) {
        logToFile(Actions.USER_ERROR, directory, 'Path is missing (user %s)', req.params.userId);
        ErrorHelper.sendNoPath(res);
      } else if (err instanceof PatchOperations.PatchNoValueError) {
        logToFile(Actions.USER_ERROR, directory, 'No value provided (userId: %s)', req.params.userId);
        ErrorHelper.sendNoValue(res);
      } else if (err instanceof PatchOperations.PatchOpMissingError) {
        logToFile(Actions.USER_ERROR, directory, 'Missing operations (userId: %s)', req.params.userId);
        ErrorHelper.sendPatchOperationsMissing(res);
      } else if (err instanceof PatchOperations.PatchInvalidFilterError) {
        logToFile(Actions.USER_ERROR, directory, 'Invalid filter (userId: %s)', req.params.userId);
        ErrorHelper.sendInvalidFilterError(res);
      } else if (err instanceof PatchOperations.PatchNoTargetError) {
        logToFile(Actions.USER_ERROR, directory, 'No path provided (userId: %s)', req.params.userId);
        ErrorHelper.sendNoPathReplace(res);
      } else if (err instanceof PatchOperations.PatchRemoveComplexAttributeError) {
        logToFile(Actions.USER_ERROR, directory, 'Tried to remove an attribute with complex filter (userId: %s)', req.params.userId);
        ErrorHelper.sendPatchRemoveComplexAttributeError(res);
      } else if (err.statusCode) {
        ErrorHelper.handleWorkflowServiceErrors(err, res, Actions.USER_ERROR, directory);
      } else {
        logToFile(Actions.USER_ERROR, directory, 'Unexpected error: %s (userId: %s)', err.message, req.params.userId);
        ErrorHelper.sendInternalServerError(res);
      }
    });
});

/**
 * DELETE /Users/:userId
 * @desc Deletes the user associated with {userId}.
 * @param {userId} the id of the user.
 */
router.delete('/:userId',
  (req, res) => {
    const userId = req.params.userId;
    const baseUrl = req.getBaseUrl();
    const directory = new Directory(baseUrl, req.authToken);

    logToFile(Actions.USER_DELETE, directory, 'DELETE user (userId: %s)', userId);

    if (typeof userId === 'undefined' || userId === '') {
      logToFile(Actions.USER_ERROR, directory, 'Id missing');
      return ErrorHelper.sendIdMissing(res);
    }

    directory
      .deleteUser(userId)
      .then(() => {
        logToFile(Actions.USER_DELETE, directory, 'DELETE user: completed (userId: %s)', userId);
        res.sendStatus(204);
      })
      .catch(err => {
        if (err.statusCode) {
          ErrorHelper.handleWorkflowServiceErrors(err, res, Actions.USER_ERROR, directory);
        } else {
          logToFile(Actions.USER_ERROR, directory, 'Unexpected error: %s (userId: %s)', err.message, userId);
          ErrorHelper.sendInternalServerError(res, err);
        }
      });
  });

module.exports = router;
