/**
 * directory.js
 * Contains the logic of a directory service.
 */
'use strict';
const user = require('./user');
const group = require('./group');
const patchoperations = require('./patchoperations');
const SearchOperation = require('../models/search');
const UserTranslator = require('./translator/usertranslator');
const GroupTranslator = require('./translator/grouptranslator');
const inject = require('../utils/injector').inject;
const intersect = require('../utils/array').intersect;
const mem = require('mem');
const config = require('../config');

const UserResource = user.UserResource;
const GroupResource = group.GroupResource;
const PatchOperationFactory = patchoperations.PatchOperationFactory;
const PatchOpMissingError = patchoperations.PatchOpMissingError;

/**
 * Represents an error thrown by the directory.
 */
class DirectoryError extends Error {
  /**
   * Instanciates a new directory error.
   * @param {String} [message] A human readable message.
   * @param {Number} [statusCode=400] The status code of the error.
   */
  constructor (message, statusCode = 400) {
    super(message);

    this._statusCode = statusCode;
  }

  /**
   * The status code of the error.
   * @returns {Number} A number representing the status code.
   */
  get statusCode () {
    return this._statusCode;
  }
}

/**
 * Represents a directory service.
 * @class Directory
 */
class Directory {
  /**
   * Creates an instance of Directory.
   * @param {String} baseUrl The base url treated by the directory service.
   * @param {String} token The token found in the request.
   * @param {Object} workflowService The injected workflow service class.
   * @memberOf Directory
   */
  constructor (baseUrl, token, workflowService = null) {
    this.baseUrl = baseUrl;
    this._workflowService = workflowService;
    this._wfgOperationContext = {
      scimToken: token,
      operationCount: 0,
      operationErrorCount: 0
    };

    if (config.app.cacheEnabled === true) {
      this.getInfo = mem(this.getInfo, {
        maxAge: config.app.cacheMaxAge
      });
    }
  }

  /**
   * Creates a new user from predefined attributes.
   * @param {UserResource} userAttributes User attributes for the new user.
   * @return {Promise<UserResource>} A promise.
   * @memberOf Directory
   */
  createUser (userAttributes) {
    let user = new UserResource(userAttributes);
    const wfgUser = new UserTranslator(UserTranslator.SCIM_TO_WFG_TYPE, user).translate();

    Object.cleanProperties(user);
    Object.cleanProperties(wfgUser);
    return this.getInfo()
      .then(directoryInfo => {
        wfgUser.directoryId = directoryInfo.id;
        return this._workflowService.invoke('createUser', this._wfgOperationContext, {
          input: wfgUser
        });
      })
      .then(result => {
        user.id = result.userId;
        user.addUrlLocation(this.baseUrl);
        return Promise.resolve(user);
      });
  }

  /**
   * Creates a groups from its attributes.
   * @param {Object} groupAttributes The attributes to create the group from.
   * @return {Promise<GroupResource>} A promise.
   * @memberOf Directory
   */
  createGroup (groupAttributes) {
    let group = new GroupResource(groupAttributes);
    const wfgGroup = new GroupTranslator(GroupTranslator.SCIM_TO_WFG_TYPE, group).translate();
    let resultedGroup;

    Object.cleanProperties(group);
    Object.cleanProperties(wfgGroup);
    return this.getInfo()
      .then(directoryInfo => {
        wfgGroup.directoryId = directoryInfo.id;
        return this._workflowService.invoke('createGroup', this._wfgOperationContext, {
          input: wfgGroup
        });
      })
      .then(result => {
        resultedGroup = result;

        if (group.members) {
          return this._workflowService.invoke('addUsersToGroup', this._wfgOperationContext, {
            input: {
              id: result.groupId,
              userIds: group.members.map(member => member.value)
            }
          });
        }

        return Promise.resolve();
      })
      .then(() => {
        group.id = resultedGroup.groupId;
        group.addUrlLocation(this.baseUrl);
        return Promise.resolve(group);
      });
  }

  /**
   * Retrieve an existing user with the option to specify the attributes
   * needed.
   * @param {String} userId The id of the existing user to retrieve.
   * @param {Array<String>} [attributes=null] A list of attributes that will be included in the resource.
   * @param {Array<String>} [excludedAttributes=null] A list of attributes that will be excluded from the resource.
   * @return {Promise<UserResource>} A promise.
   * @memberOf Directory
   */
  async getUser (userId, attributes = null, excludedAttributes = null) {
    const [directoryInfo, result] = await Promise.all([
      this.getInfo(),
      this._workflowService.invoke('getUser', this._wfgOperationContext, {id: userId})
    ]);

    if (directoryInfo.id !== result.directoryId) {
      throw new DirectoryError('The user does not exists in this directory.', 404);
    }

    const translator = new UserTranslator(UserTranslator.WFG_TO_SCIM_TYPE, result);
    let user = translator.translate();

    if (attributes) {
      user.includesOnly(...attributes);
    } else if (excludedAttributes) {
      user.excludes(...excludedAttributes);
    }

    user.checkAddBusinessExtension();
    Object.cleanProperties(user);
    user.addUrlLocation(this.baseUrl);
    return user;
  }

  /**
   * Retrieve an existing group with the option to specify the attributes.
   * needed.
   * @param {String} groupId The id of the group to retrieve.
   * @param {Array<String>} [attributes=null] A list of attributes to be included in the resource.
   * @param {Array<String>} [excludedAttributes=null] A list of attributes to be excluded from the resource.
   * @return {Promise<GroupResource>} A promise that resolves with the group.
   * @memberOf Directory
   */
  async getGroup (groupId, attributes = null, excludedAttributes = null) {
    const [directoryInfo, wfgGroup] = await Promise.all([
      this.getInfo(),
      this._workflowService.invoke('getGroup', this._wfgOperationContext, {
        id: groupId
      })
    ]);

    if (directoryInfo.id !== wfgGroup.directoryId) {
      throw new DirectoryError('The requested group does not exists in this directory.', 404);
    }

    const wfgUsers = await this._workflowService.invoke('getUsers', this._wfgOperationContext, {
      groupId: wfgGroup.id
    });
    const translator = new GroupTranslator(GroupTranslator.WFG_TO_SCIM_TYPE, wfgGroup);
    const group = translator.translate();

    group.members = wfgUsers.items
      .map(user => new UserTranslator(UserTranslator.WFG_TO_SCIM_TYPE, user).translate())
      .map(user => {
        user.addUrlLocation(this.baseUrl);
        return {
          value: user.id,
          display: user.displayName,
          '$ref': user.meta.location
        };
      });

    if (attributes) {
      group.includesOnly(...attributes);
    } else if (excludedAttributes) {
      group.excludes(...excludedAttributes);
    }

    Object.cleanProperties(group);
    group.addUrlLocation(this.baseUrl);
    return group;
  }

  /**
   * @description Modifies a user accordingly to specified operations.
   * @param {String} userId The id of an existing user.
   * @param {Array<Object>} operationsToExecute An array of operations to execute on the user.
   * @return {Promise<UserResource>} A promise. UserResource if resolved.
   * @memberOf Directory
   */
  async partialUpdateUser (userId, operationsToExecute) {
    const [directoryInfo, wfgUser] = await Promise.all([
      this.getInfo(),
      this._workflowService.invoke('getUser', this._wfgOperationContext, {
        id: userId
      })
    ]);

    if (directoryInfo.id !== wfgUser.directoryId) {
      throw new DirectoryError('The user cannot be found in this directory.', 404);
    }

    let translator = new UserTranslator(UserTranslator.WFG_TO_SCIM_TYPE, wfgUser);
    let scimUser = translator.translate();

    for (let op of operationsToExecute) {
      if (!op.op) {
        throw new PatchOpMissingError();
      }

      let operation = PatchOperationFactory.fromOperation(op);

      scimUser = operation.execute(scimUser);
    }

    scimUser.checkAddBusinessExtension();
    translator = new UserTranslator(UserTranslator.SCIM_TO_WFG_TYPE, scimUser);
    await this._workflowService.invoke('updateUser', this._wfgOperationContext, {
      input: translator.translate()
    });

    Object.cleanProperties(scimUser);
    scimUser.addUrlLocation(this.baseUrl);
    return scimUser;
  }

  /**
   * @description Modifies a group according to specified operations.
   * @param {String} groupId The id of an existing group.
   * @param {Array<Object>} operationsToExecute An array of operations to execute on the group.
   * @return {Promise<GroupResource>} A promise. GroupResource if resolved.
   * @memberOf Directory
   */
  partialUpdateGroup (groupId, operationsToExecute) {
    let
      /** @type {GroupResource} */ updatedGroup;

    return this.getGroup(groupId)
      .then(group => {
        const memberIds = group.members ? group.members.map(m => m.value) : [];

        for (let op of operationsToExecute) {
          if (!op.op) {
            throw new PatchOpMissingError();
          }

          const operation = PatchOperationFactory.fromOperation(op);

          try {
            group = operation.execute(group);
          } catch (err) {
            return Promise.reject(err);
          }

          group.members = [].concat.apply([], group.members);

          let keys = {};
          group.members = group.members.filter(m => {
            const seen = keys[m.value];
            keys[m.value] = true;
            return typeof seen === 'undefined';
          });
        }

        updatedGroup = group;
        const newMembersIds = group.members.map(m => m.value);
        const commonIds = intersect(memberIds, newMembersIds);
        const toDeleteIds = memberIds.filter(id => !commonIds.includes(id));
        const toAddIds = newMembersIds.filter(id => !commonIds.includes(id));
        const newGroup = new GroupTranslator(GroupTranslator.SCIM_TO_WFG_TYPE, group)
          .translate();
        delete newGroup.users;

        return Promise.all([
          this._workflowService.invoke('updateGroup', this._wfgOperationContext, {
            input: newGroup
          }),
          toAddIds.length > 0
            ? this._workflowService.invoke('addUsersToGroup', this._wfgOperationContext, {
              input: {
                id: group.id,
                userIds: toAddIds
              }
            })
            : null,
          toDeleteIds.length > 0
            ? this._workflowService.invoke('removeUsersFromGroup', this._wfgOperationContext, {
              input: {
                id: group.id,
                userIds: toDeleteIds
              }
            })
            : null
        ]);
      })
      .then(([updateRes]) => this._workflowService.invoke('getUsers', this._wfgOperationContext, {
        groupId: updateRes.groupId
      }))
      .then(usersRes => {
        updatedGroup.members = usersRes.items
          .map(u => new UserTranslator(UserTranslator.WFG_TO_SCIM_TYPE, u).translate())
          .map(u => {
            u.addUrlLocation(this.baseUrl);
            return {
              value: u.id,
              display: u.displayName,
              '$ref': u.meta.location
            };
          });

        Object.cleanProperties(updatedGroup);
        updatedGroup.addUrlLocation(this.baseUrl);
        return Promise.resolve(updatedGroup);
      });
  }

  /**
   * @description Deletes a user from the directory.
   * @param {String} userId The id of an existing user.
   * @return {Promise} A promise.
   * @memberOf Directory
   */
  deleteUser (userId) {
    return this._workflowService.invoke('deleteUser', this._wfgOperationContext, {
      input: {
        id: userId
      }
    });
  }

  /**
   * @desc Deletes a group from the directory.
   * @param {String} groupId The id of the group.
   * @return {Promise} A promise.
   */
  deleteGroup (groupId) {
    return this._workflowService.invoke('deleteGroup', this._wfgOperationContext, {
      input: {
        id: groupId
      }
    });
  }

  /**
   * Search in the directory for resource according to parameters.
   * @param {String} [schema=null] The schema to restrict directory search.
   * @param {String} [filter=null] The search filter string not parsed.
   * @param {String} [sortBy=null] The attribute name from which to sort by.
   * @param {String} [sortOrder=null] ASCENDING | DESCENDING. The order of the sort.
   * @param {number} [startIndex=0] The start index of the pagination.
   * @param {number} [count=0] The maximum number of results to return.
   * @param {Array<String>} [attributes=null] The attributes to include.
   * @param {Array<String>} [excludedAttributes=null] The attributes to exclude.
   * @return {Promise<Array<Resource>>} A promise with the results.
   * @memberOf Directory
   */
  async search (
    schema = null,
    filter = null,
    sortBy = null,
    sortOrder = null,
    startIndex = 0,
    count = 0,
    attributes = null,
    excludedAttributes = null
  ) {
    const search = SearchOperation.Search.fromSchema(schema);

    search.context = this._wfgOperationContext;
    search.directoryId = (await this.getInfo()).id;

    if (filter) {
      search.filter = filter;
    }

    if (sortBy) {
      search.sortBy = sortBy;
    }

    if (sortOrder) {
      search.sortOrder = sortOrder;
    }

    if (startIndex) {
      search.startIndex = startIndex;
    }

    if (count) {
      search.count = count;
    }

    if (attributes) {
      search.attributes = attributes;
    }

    if (excludedAttributes) {
      search.excludedAttributes = excludedAttributes;
    }

    const results = await search.execute();
    return results;
  }

  getInfo () {
    return this._workflowService.invoke('getDirectories', this._wfgOperationContext, {
      filter: {
        synchronizationConnector: 'AZURE_SCIM_V2'
      }
    })
      .then(results => {
        if (results.totalCount <= 0) {
          return Promise.resolve(null);
        }

        return Promise.resolve(results.items[0]);
      });
  }
}

module.exports = inject(Directory);
