/**
 * filter.js
 * Contains the definition of the Filter class.
 */
'use strict';

const FilterLexer = require('../utils/filter/lexer');
const FilterParser = require('../utils/filter/parser');
const Expressions = require('../utils/filter/expressions');

/**
 * Represents a SCIM filter
 * @class Filter
 */
class Filter {
  /**
   * @desc Instanciates a Filter.
   * @param {String} filterString A raw string containing the filter.
   */
  constructor (filterString) {
    this._filterString = filterString;
    this.expressionTree = new FilterParser(
      new FilterLexer(
        filterString
      )).parse();
  }

  /**
   * @desc Tests an object to know if the filter is applicable.
   * @param {Object} obj An object to test.
   * @return {Boolean} `true` if the filter can be applied. `false` otherwise.
   */
  test (obj) {
    return this.expressionTree.apply(obj);
  }

  /**
   * @desc Parses the AST to obtain a basic javascript object that can be used by the application.
   * Note that this function is very basic and doesn't get very far into the
   * parsing of the filter. It doesn't support OR expressions yet as well.
   *
   * @param {Object} tree An AST representing the filter.
   * @return {Object} A basic object that says which properties should be applied for filtering.
   */
  getJavaScriptObject (tree) {
    if (!tree) {
      tree = this.expressionTree;
    }

    // TODO: Make it recursive.

    if (tree instanceof Expressions.SimpleExpression && tree.operator === 'EQ') {
      return {
        [tree.left]: tree.right
      };
    } else if (tree instanceof Expressions.CompoundExpression) {
      let obj = Object.assign({}, this.getJavaScriptObject(tree.left));
      obj = Object.assign(obj, this.getJavaScriptObject(tree.right));
      return obj;
    }
  }
}

module.exports = Filter;
