'use strict';

let constants = require('../constants');
const { logToFile, Actions } = require('./logger');

/**
 * Contains methods to add details to HTTP errors
 * @class ErrorHelper
 */
class ErrorHelper {
  /**
   * Creates an instance of ErrorHelper.
   * @param {Number} status The error http status code.
   * @param {String} detail The detail message of the error.
   * @param {String} [scimType=null] The scim type of the error.
   * @constructor
   */
  constructor (status, detail, scimType = null) {
    this.status = status;
    this.detail = detail;

    if (scimType) {
      this.scimType = scimType;
    }

    this.schemas = [constants.schemas.error];
  }

  /**
   * Adds an error to the HTTP response according to the SCIM spec.
   * @static
   * @param {Response} res The HTTP response to add the error to.
   * @param {Number} status The HTTP status code to use.
   * @param {String} detail Human readable text explaining the error.
   * @param {String} scimType The scimType of the HTTP error. Only used for 400 errors.
   */
  static addErrorResponse (res, status, detail, scimType = null) {
    let error = {
      status: status,
      detail: detail,
      schemas: [constants.schemas.error]
    };

    if (scimType) {
      error.scimType = scimType;
    }

    res.status(status).json(error);
  }

  /**
   * Adds the 404 error code to the HTTP response.
   * @static
   * @param {Response} res The HTTP response to add the error to.
   */
  static sendErrorNotFound (res) {
    this.addErrorResponse(
      res,
      404,
      'Requested resource was not found');
  }

  /**
   * Adds an ID missing error to the response. Uses HTTP error code 400.
   * @static
   * @param {Response} res The HTTP response to add the error to.
   */
  static sendIdMissing (res) {
    this.addErrorResponse(
      res,
      400,
      'No resource ID was provided',
      'invalidValue'
    );
  }

  /**
   * Adds an incorrect schema error to the response.
   * @static
   * @param {Response} res The HTTP response to add the error to.
   */
  static sendIncorrectSchema (res) {
    this.addErrorResponse(
      res,
      400,
      'Incorrect SCIM schema in the request body',
      'invalidValue'
    );
  }

  /**
   * Adds an invalid path error for PATCH remove of a complex attribute.
   * @static
   * @param {Response} res The HTTP response to add the error to.
   */
  static sendPatchRemoveComplexAttributeError (res) {
    this.addErrorResponse(
      res,
      400,
      'Tried to remove a complex attribute.',
      'invalidPath'
    );
  }

  /**
   * Adds an operation array required error for the PATCH operation
   * @static
   * @param {Response} res The HTTP response to add the error to.
   */
  static sendPatchOperationsMissing (res) {
    this.addErrorResponse(
      res,
      400,
      'Missing the required PATCH Operations array.',
      'invalidValue'
    );
  }

  /**
   * Adds an op member required error for the PATCH operation
   * @static
   * @param {Response} res The HTTP response to add the error to.
   */
  static sendPatchOpMissing (res) {
    this.addErrorResponse(
      res,
      400,
      'Missing the required PATCH op member.',
      'invalidValue'
    );
  }

  /**
   * Adds a generic invalid value error.
   * @static
   * @param {Response} res The HTTP response to add the error to.
   * @param {String} message The error's message.
   */
  static sendInvalidValue (res, message) {
    this.addErrorResponse(
      res,
      400,
      message,
      'invalidValue'
    );
  }

  /**
   * Adds a invalid path error for a PATCH request
   * @static
   * @param {Response} res The HTTP response to add the error to.
   */
  static sendNoPath (res) {
    this.addErrorResponse(
      res,
      400,
      'Remove operation has unspecified path',
      'noTarget'
    );
  }

  /**
   * Adds a noTarget error for a PATCH replace request with filter
   * @static
   * @param {Response} res The HTTP response to add the error to.
   */
  static sendNoPathReplace (res) {
    this.addErrorResponse(
      res,
      400,
      'Replace operation filter found no element',
      'noTarget'
    );
  }

  /**
   * Adds a missing value error for a PATCH request
   * @static
   * @param {Response} res The HTTP response to add the error to.
   */
  static sendNoValue (res) {
    this.addErrorResponse(
      res,
      400,
      'No value was provided for an operation',
      'invalidValue'
    );
  }

  /**
   * Adds a mutability error for a PATCH operation
   * @static
   * @param {Response} res The HTTP response to add the error to.
   */
  static sendImmutablePatchError (res) {
    this.addErrorResponse(
      res,
      400,
      'Patch operation tried to modify an immutable attribute.',
      'mutability'
    );
  }

  /**
   * Adds a unauthorized error for a request.
   * @static
   * @param {Response} res The HTTP response to add the error to.
   */
  static sendUnauthorizedError (res) {
    this.addErrorResponse(
      res,
      401,
      'This operation is not authorized on this resource.'
    );
  }

  /**
   * Sends the SCIM error relating to uniqueness of a resource.
   * @static
   * @param {Response} res the HTTP response to add the error to.
   */
  static sendUniquenessError (res) {
    this.addErrorResponse(
      res,
      409,
      'Element exists in the database',
      'uniqueness'
    );
  }

  /**
   * Adds an internal server error to the response (HTTP 500).
   * @static
   * @param {Response} res The HTTP response to add the error to.
   * @param {String} [detail=null] The detail of the error.
   */
  static sendInternalServerError (res, detail = null) {
    this.addErrorResponse(
      res,
      500,
      detail === null ? 'Internal server error' : detail
    );
  }

  /**
   * Adds an invalid syntax error (one or more attributes missing)
   * @static
   * @param {Response} res The HTTP response to add the error to.
   */
  static sendInvalidSyntax (res) {
    this.addErrorResponse(
      res,
      400,
      'One or more required attributes were not provided.',
      'invalidSyntax'
    );
  }

  /**
   * Adds an invalid filter error.
   * @static
   * @param {Response} res The HTTP response to add the error to.
   */
  static sendInvalidFilterError (res) {
    this.addErrorResponse(
      res,
      400,
      'The filter provided is invalid.',
      'invalidFilter'
    );
  }

  /**
   * @desc Handles errors returned or thrown by the Workflow Service.
   * @param {Object} err The error from the workflow service.
   * @param {Object} res The express response object.
   * @param {String} [logAction=null] The log action to be passed to the logger.
   * @param {Object} [directory=null] The directory object that performs the action.
   */
  static handleWorkflowServiceErrors (err, res, logAction = null, directory = null) {
    switch (err.statusCode) {
      case 400:
        if (
          err.message.indexOf('ID') !== -1 &&
          err.message.indexOf('invalid') !== -1
        ) {
          logToFile(logAction, directory, 'Not found');
          return this.sendErrorNotFound(res);
        }

        logToFile(logAction, directory, 'Invalid value');
        return this.sendInvalidValue(res, err.message);

      case 404:
        logToFile(logAction, directory, 'Resource not found in the directory.');
        return this.sendErrorNotFound(res);

      case 409:
        logToFile(logAction, directory, 'Already exists');
        return this.sendUniquenessError(res);

      case 401:
      case 403:
        logToFile(Actions.GENERAL, directory, 'Unauthorized access attempt');
        return this.sendUnauthorizedError(res);

      case 500:
        if (err.message.indexOf('does not exist') !== -1) {
          logToFile(logAction, directory, 'Not found');
          return this.sendErrorNotFound(res);
        }

        if (err.message.indexOf('already exists') !== -1) {
          logToFile(logAction, directory, 'Already exists');
          return ErrorHelper.sendUniquenessError(res);
        }

        logToFile(logAction, directory, 'Unexpected error: "%s"', err.message);
        return this.sendInternalServerError(res, err.message);

      default:
        break;
    }
  }
}

module.exports = ErrorHelper;
