/**
 * patchoperations.js
 * Contains logic to implement the SCIM PATCH command.
 */
'use strict';

const FilterLexer = require('../utils/filter/lexer');
const FilterParser = require('../utils/filter/parser');
const Operators = require('../utils/filter/operators');
const constants = require('../models/constants');

const ADD_OPERATION = 'add';
const REPLACE_OPERATION = 'replace';
const REMOVE_OPERATION = 'remove';
const BOOLEAN_PROPS = new Set([
  'active',
  'primary'
]);

/**
 * General class for PATCH errors.
 * @class PatchOperationError
 * @extends {Error}
 */
class PatchOperationError extends Error { }

/**
 * Error whereby no value was provided for a PATCH operation.
 * @class PatchNoValueError
 * @extends {PatchOperationError}
 */
class PatchNoValueError extends PatchOperationError {
  /**
   * Creates an instance of PatchNoValueError.
   */
  constructor () {
    super('No value attribute was provided.');
  }
}

/**
 * Error whereby a target was not provided for a PATCH operation.
 * @class PatchNoPathError
 * @extends {PatchOperationError}
 */
class PatchNoPathError extends PatchOperationError {
  /**
   * Creates an instance of PatchNoPathError.
   */
  constructor () {
    super('No path attribute was provided.');
  }
}

/**
 * Error whereby an operation tried to modify an immutable attribute
 * @class PatchNoPathError
 * @extends {PatchOperationError}
 */
class PatchImmutableError extends PatchOperationError {
  /**
   * Creates an instance of PatchNoPathError.
   */
  constructor () {
    super('PATCH operations cannot replace an existing immutable attribute.');
  }
}

/**
 * Error thrown when an operation could not be identified or parsed.
 * @class PatchOpMissingError
 * @extends {PatchOperationError}
 */
class PatchOpMissingError extends PatchOperationError {
  /**
   * Creates an instance of PatchOpMissingError.
   */
  constructor () {
    super('No patch operations found');
  }
}

/**
 * Error thrown when trying to remove a complex attribute.
 * @class PatchRemoveComplexAttributeError
 * @extends {PatchOperationError}
 */
class PatchRemoveComplexAttributeError extends PatchOperationError {
  /**
   * Creates an instance of PatchRemoveComplexAttributeError.
   */
  constructor () {
    super('Cannot remove a complex attribute.');
  }
}

/**
 * Error thrown when a path filter is invalid.
 * @class PatchInvalidFilterError
 * @extends {PatchOperationError}
 */
class PatchInvalidFilterError extends PatchOperationError {
  /**
   * Creates an instance of PatchInvalidFilterError.
   */
  constructor () {
    super('Path filter is invalid');
  }
}

/**
 * Error thrown when a path filter finds no element.
 * @class PatchNoTargetError
 * @extends {PatchOperationError}
 */
class PatchNoTargetError extends PatchOperationError {
  /**
   * Creates an instance of PatchNoTargetError.
   */
  constructor () {
    super('Path filter found no element');
  }
}

/**
 * General class for all PATCH operations.
 * @class PatchOperation
 */
class PatchOperation {
  /**
   * Creates an instance of PatchOperation.
   * @param {Object} operation The operation to execute (corresponds to the SCIM spec).
   */
  constructor (operation) {
    this.operation = operation.op.toLowerCase().trim();
    this.fullPath = operation.path;
    this.path = [];
    this.value = operation.value;

    if (!operation.path) {
      return;
    }

    let filterIndex;
    if ((filterIndex = operation.path.indexOf('[')) === -1) {
      this.path = this.path.concat(
        operation.path.split('.')
      );
    } else {
      this.path = this.path.concat(
        operation.path.substring(
          0,
          filterIndex
        ).split('.')
      );

      let filteredPath = operation.path.substring(filterIndex).split(']');
      let filter = filteredPath[0].replace(/[[\]]/g, '');

      if (filteredPath.length > 1 && filteredPath[1] !== '') {
        // Supplementary path after the filter (minus the .)
        this.attributePath = filteredPath[1].substring(1);
      }

      let lexer;

      try {
        lexer = new FilterLexer(filter);
        this.filter = new FilterParser(lexer).parse();
      } catch (_) {
        throw new PatchInvalidFilterError();
      }
    }

    if (
      this.fullPath.includes('members') &&
      (this.attributePath || this.path[this.path.length - 1] !== 'members')
    ) {
      // Data on members when patching groups is immutable
      throw new PatchImmutableError();
    }

    if (this.path[this.path.length - 1] !== 'members' && !this.attributePath) {
      // Support for azure
      if (
        this.value instanceof Array &&
        this.value.length > 0 &&
        this.value[0].hasOwnProperty('$ref')
      ) {
        this.value = this.value[0].value;
      } else if (
        typeof this.value === 'object' &&
        this.value.hasOwnProperty('$ref')
      ) {
        this.value = this.value.value;
      }
    }
  }

  /**
   * Sets a value in a deeply nested object according to a path.
   * @param {Object} obj The object to modify.
   * @param {Object} value The value to set.
   * @return {Object} The modified object.
   */
  setProp (obj, value) {
    let len = this.path.length;
    for (var i = 0; i < len - 1; i++) {
      let elem = this.path[i];
      if (!obj[elem]) {
        obj[elem] = {};
      }
      obj = obj[elem];
    }

    let propToSet = this.path[len - 1];
    if (BOOLEAN_PROPS.has(propToSet) && typeof value === 'string') {
      obj[this.path[len - 1]] = value === 'True';
    } else {
      obj[this.path[len - 1]] = value;
    }

    return obj;
  }

  /**
   * Sets the information needed to execute a PATCH operation.
   * @param {Object} obj The object to modify.
   */
  execute (obj) {
    let objToModify = obj;

    if (this.path && this.path.length > 0) {
      if (this.path.includes('manager')) {
        if (!objToModify.hasOwnProperty(constants.schemas.extensions.user.business)) {
          objToModify[constants.schemas.extensions.user.business] = {};
        } else if (!objToModify.hasOwnProperty('manager')) {
          objToModify.manager = {};
        }

        if (!this.path.includes('id')) {
          this.path.push('id');
        }

        this.path = [constants.schemas.extensions.user.business]
          .concat(this.path);
      }

      for (let prop of this.path) {
        if (objToModify) {
          objToModify = objToModify[prop];
        }
      }
    }

    this.objToModify = objToModify;
    this.isMultiValued = objToModify instanceof Array;
    this.isComplex = objToModify instanceof Object;
  }
}

/**
 * Contains logic to apply a PATCH add operation.
 * @class PatchAddOperation
 * @extends {PatchOperation}
 */
class PatchAddOperation extends PatchOperation {
  /**
   * Creates an instance of PatchAddOperation.
   * @param {Object} operation The operation to execute.
   */
  constructor (operation) {
    super(operation);

    this.noChanges = true;

    if (this.operation !== ADD_OPERATION) {
      throw new TypeError('Only add is supported');
    }
  }

  /**
   * Executes a PATCH add operation on an object.
   * @param {Object} obj The object to modify.
   * @return {Object} The modified object.
   */
  execute (obj) {
    super.execute(obj);

    if (typeof this.value === 'undefined' || this.value === null) {
      throw new PatchNoValueError();
    }

    if (this.isMultiValued) {
      // Set all other primary in the Array to false if it was set
      if (this.value.primary) {
        for (let elem of this.objToModify) {
          elem.primary = false;
        }
      }

      // Add the new array member
      this.objToModify.push(this.value);
      this.noChanges = false;
    } else if (this.isComplex && this.value instanceof Object) {
      // Add all attributes to the complex object
      for (let prop in this.value) {
        if (prop) {
          if (this.objToModify[prop] !== this.value[prop]) {
            this.noChanges = false;
          }

          this.objToModify[prop] = this.value[prop];
        }
      }
    } else if (this.isComplex) {
      switch (true) {
        case this.objToModify.hasOwnProperty('id'):
          this.objToModify.id = this.value;
          break;

        case this.objToModify.hasOwnProperty('value'):
          this.objToModify.value = this.value;
          break;

        default:
          for (let prop in this.objToModify) {
            this.objToModify[prop] = this.value;
          }
          break;
      }
    } else {
      if (this.value !== this.objToModify) {
        this.noChanges = false;
      }

      this.setProp(obj, this.value);
    }

    return obj;
  }
}

/**
 * Contains logic to execute a PATCH remove operation.
 * @class PatchRemoveOperation
 * @extends {PatchOperation}
 */
class PatchRemoveOperation extends PatchOperation {
  /**
   * Creates an instance of PatchRemoveOperation.
   * @param {Object} operation The operation to execute.
   */
  constructor (operation) {
    super(operation);

    if (this.operation !== REMOVE_OPERATION) {
      throw new TypeError('Only remove is supported');
    }
  }

  /**
   * Executes a PATCH remove operation on an object.
   * @param {Object} obj The object to modify.
   * @return {Object} The modified object.
   */
  execute (obj) {
    if (!this.fullPath || this.fullPath === '') {
      throw new PatchNoPathError();
    }

    super.execute(obj);

    if (this.isMultiValued) {
      if (this.filter) {
        this.setProp(
          obj,
          this.objToModify.filter(
            elem => {
              try {
                return !this.filter.apply(elem);
              } catch (err) {
                if (err instanceof Operators.UnsupportedOperationError) {
                  throw new PatchInvalidFilterError();
                } else {
                  throw err;
                }
              }
            }
          )
        );
      } else if (this.fullPath.endsWith('members') &&
        this.value &&
        this.value instanceof Array &&
        this.value.length > 0 &&
        this.value[0].value &&
        this.value[0].$ref === null
      ) {
        // Azure sends the members to remove instead of using a filter
        for (let removedMember of this.value) {
          this.setProp(obj, this.objToModify.filter(
            member => member.value !== removedMember.value
          ));
        }
      } else {
        this.setProp(obj, []);
      }
    } else if (this.isComplex) {
      throw new PatchRemoveComplexAttributeError();
    } else {
      if (this.fullPath === 'userName') {
        throw new PatchImmutableError();
      }
      this.setProp(obj, null);
    }

    return obj;
  }
}

/**
 * Contains logic to execute a PATCH replace operation.
 * @class PatchReplaceOperation
 * @extends {PatchOperation}
 */
class PatchReplaceOperation extends PatchOperation {
  /**
   * Creates an instance of PatchReplaceOperation.
   * @param {Object} operation The operation to execute.
   */
  constructor (operation) {
    super(operation);

    if (this.operation !== REPLACE_OPERATION) {
      throw new TypeError('Only replace is supported');
    }
  }

  /**
   * Executes a PATCH replace operation on an object.
   * @param {Object} obj The object to modify.
   * @return {Object} The modified object.
   */
  execute (obj) {
    super.execute(obj);

    if (this.isMultiValued) {
      if (
        this.value.primary ||
        (this.attributePath === 'primary' && this.value)
      ) {
        // Set all other primaries
        for (let elem of this.objToModify) {
          elem.primary = false;
        }
      }

      if (this.filter) {
        // Replace each value that matches the filter
        let replacedArray = [];
        let modified = 0;

        this.objToModify.forEach(elem => {
          let matches = false;

          try {
            matches = this.filter.apply(elem);
          } catch (err) {
            if (err instanceof Operators.UnsupportedOperationError) {
              throw new PatchInvalidFilterError();
            } else {
              throw err;
            }
          }

          if (matches) {
            if (this.attributePath) {
              // Set a single attribute
              elem[this.attributePath] = this.value;
              replacedArray.push(elem);
            } else {
              // Replace the whole object
              replacedArray.push(this.value);
            }
            modified++;
          } else {
            replacedArray.push(elem);
          }
        });

        if (modified === 0) {
          throw new PatchNoTargetError();
        }

        this.setProp(obj, replacedArray);
      } else {
        // Else replace the array
        if (this.objToModify.length === 0 && typeof this.value === 'object') {
          this.setProp(obj, Array(this.objToModify.length + 1).fill(this.value));
        } else {
          this.setProp(obj, Array(this.objToModify.length).fill(this.value));
        }
      }
    } else if (this.isComplex && this.value instanceof Object) {
      for (let prop in this.value) {
        if (prop) {
          this.objToModify[prop] = this.value[prop];
        }
      }
    } else if (this.isComplex) {
      switch (true) {
        case this.objToModify.hasOwnProperty('id'):
          this.objToModify.id = this.value;
          break;

        case this.objToModify.hasOwnProperty('value'):
          this.objToModify.value = this.value;
          break;

        default:
          for (let prop in this.objToModify) {
            this.objToModify[prop] = this.value;
          }
          break;
      }
    } else {
      this.setProp(obj, this.value);
    }

    return obj;
  }
}

/**
 * Factory for PATCH operations.
 * @class PatchOperationFactory
 */
class PatchOperationFactory {
  /**
   * Factory to create PATCH operations.
   * @static
   * @param {Object} operation The operation to create.
   * @return {PatchOperation} A implementation of a PATCH operation.
   */
  static fromOperation (operation) {
    let operationType = operation.op.toLowerCase().trim();
    switch (operationType) {
      case ADD_OPERATION:
        return new PatchAddOperation(operation);
      case REMOVE_OPERATION:
        return new PatchRemoveOperation(operation);
      case REPLACE_OPERATION:
        return new PatchReplaceOperation(operation);
      default:
        throw new TypeError('This operation is not supported');
    }
  }
}

module.exports = {
  PatchOperationError,
  PatchNoPathError,
  PatchNoValueError,
  PatchImmutableError,
  PatchOpMissingError,
  PatchRemoveComplexAttributeError,
  PatchInvalidFilterError,
  PatchNoTargetError,
  PatchOperation,
  PatchAddOperation,
  PatchRemoveOperation,
  PatchReplaceOperation,
  PatchOperationFactory,
  ADD_OPERATION,
  REPLACE_OPERATION,
  REMOVE_OPERATION
};
