/**
 * resource.js
 * Shared model for SCIM resources.
 */
'use strict';

/**
 * Abstract. Represents a SCIM resource.
 * @class Resource
 */
class Resource {
  /**
   * Creates an instance of Resource.
   */
  constructor (resourceType) {
    this._requiredAttributes = [
      'id',
      'schemas',
      'meta',
      '_requiredAttributes'
    ];

    this.meta = {
      resourceType
    };
  }

  /**
   * @description Deletes properties that are not needed.
   * @param {Array} attributes Attributes that will only appear in the representation.
   * @return {Object} Return this instance.
   */
  includesOnly (...attributes) {
    if (attributes.length > 0) {
      for (let prop of Object.keys(this)) {
        if (
          !attributes.includes(prop) &&
          this._requiredAttributes.every(element => element !== prop)
        ) {
          delete this[prop];
        }
      }
    }

    return this;
  }

  /**
   * @description Deletes the excluded properties from this object.
   * @param {Array} attributes Attributes to be excluded.
   * @return {Object} Returns this instance.
   */
  excludes (...attributes) {
    for (let attr of attributes) {
      if (
        this.hasOwnProperty(attr) &&
        this._requiredAttributes.every(element => element !== attr)
      ) {
        delete this[attr];
      }
    }

    return this;
  }

  /**
   * Overrides json serialization process.
   * @return {Resource} this object but clean.
   */
  toJSON () {
    delete this._requiredAttributes;
    Object.cleanProperties(this);
    return this;
  }
}

module.exports = Resource;
