/**
 * search.js
 * Model for the SCIM search options.
 */
'use strict';

const Filter = require('./filter');
const constants = require('../models/constants');
const UserTranslator = require('./translator/usertranslator');
const GroupTranslator = require('./translator/grouptranslator');
const inject = require('../utils/injector').inject;

const ASCENDING = 'ascending';
const DESCENDING = 'descending';

class TooManyResultsError extends Error { }
class SearchError extends Error {}
class InvalidSortOrderError extends SearchError {
  constructor () {
    super('The sort order is invalid. It should be ASCENDING or DESCENDING');
  }
}
class InvalidFilterError extends SearchError {}
class InvalidStartIndexError extends SearchError {
  constructor () {
    super('The start index is not a number.');
  }
}
class InvalidCountError extends SearchError {
  constructor () {
    super('The count is not a number.');
  }
}
class UnauthorizedSearchError extends SearchError {
  constructor () {
    super('Unauthorized search on this resource.');
  }
}

/**
 * SCIM Search model.
 * @class Search
 */
class Search {
  /**
   * Creates an instance of Search.
   */
  constructor (workflowService = null) {
    this._filter = null;
    this._resourceSchema = null;
    this._sortBy = null;
    this._sortOrder = null;
    this._startIndex = 0;
    this._count = 0;
    this._attributes = null;
    this._excludedAttributes = null;
    this._searchUsers = false;
    this._searchGroups = false;
    this._workflowService = workflowService;
    this._wfgOperationContext = null;
    this._directoryId = '';
  }

  set directoryId (id) {
    this._directoryId = id;
  }

  set context (context) {
    this._wfgOperationContext = context;
  }

  set searchUsers (value) {
    this._searchUsers = value;
  }

  set searchGroups (value) {
    this._searchGroups = value;
  }

  set filter (value) {
    try {
      this._filter = new Filter(value);
    } catch (err) {
      throw new InvalidFilterError();
    }
  }

  set resourceSchema (value) {
    this._resourceSchema = value;
  }

  set sortBy (value) {
    this._sortBy = value;
  }

  set sortOrder (value) {
    if (value !== ASCENDING && value !== DESCENDING) {
      throw new InvalidSortOrderError();
    } else {
      this._sortOrder = value;
    }
  }

  set startIndex (value) {
    let max = Math.max(value, 1);

    if (isNaN(max)) {
      throw new InvalidStartIndexError();
    }

    this._startIndex = max;
  }

  get startIndex () {
    return this._startIndex;
  }

  get count () {
    return this._count;
  }

  set count (value) {
    let max = Math.max(value, 0);

    if (isNaN(max)) {
      throw new InvalidCountError();
    }

    this._count = max;
  }

  set attributes (value) {
    this._attributes = value;
  }

  set excludedAttributes (value) {
    this._excludedAttributes = value;
  }

  /**
   * Creates an instance of Search with a schema.
   * @param {String} resourceSchema The SCIM schema to search on.
   * @return {Search} The search model.
   */
  static fromSchema (resourceSchema, /* WorkflowService */ workflowService) {
    let search = new Search(workflowService);
    search.resourceSchema = resourceSchema;

    if (resourceSchema === constants.schemas.user) {
      search.searchUsers = true;
    }

    if (resourceSchema === constants.schemas.group) {
      search.searchGroups = true;
    }

    return search;
  }

  /**
   * Searches user resources.
   * @param {Function} callback The function to call back with the results.
   */
  _search (callback) {
    let filter = {};
    let additionalProcessing = (results) => {
      return results;
    };
    let
      /** @type {Translator} */ translator,
      /** @type {String} */ methodName,
      /** @type {Class} */ TranslatorClass;

    if (this._searchUsers && !this._searchGroups) {
      translator = new UserTranslator(UserTranslator.SCIM_TO_WFG_TYPE);
      methodName = 'getUsers';
      TranslatorClass = UserTranslator;

      if (this._filter) {
        try {
          filter = this._filter.getJavaScriptObject();
        } catch (e) {
          throw new InvalidFilterError();
        }
      }
    } else if (this._searchGroups && !this._searchUsers) {
      translator = new GroupTranslator(GroupTranslator.SCIM_TO_WFG_TYPE);
      methodName = 'getGroups';
      TranslatorClass = GroupTranslator;

      if (this._filter) {
        try {
          filter = this._filter.getJavaScriptObject();
        } catch (e) {
          throw new InvalidFilterError();
        }
      }
    }

    if (filter) {
      // Transform the SCIM attribute into a WFG attribute
      for (let prop in filter) {
        const translatedProp = translator.translateAttribute(prop);

        if (translatedProp !== prop) {
          filter[translatedProp] = filter[prop];
          delete filter[prop];
        }
      }

      if (filter.hasOwnProperty('userName')) {
        filter.userNames = [filter.userName];
        delete filter.userName;
      }

      if (filter.hasOwnProperty('id') && filter.hasOwnProperty('managerId')) {
        filter.ids = [filter.id];
        additionalProcessing = function (results) {
          results.items = results.items.filter(item => item.managerId === this.managerId);
          return results;
        }.bind({managerId: filter.managerId});

        delete filter.id;
        delete filter.managerId;
      }

      if (filter.hasOwnProperty('id') && filter.hasOwnProperty('users')) {
        filter.ids = [filter.id];
        additionalProcessing = function (results) {
          return new Promise(function (resolve, reject) {
            results.items = results.items.filter(g => g.id === this.groupId);

            if (results.items.length > 0) {
              this.searchThis._workflowService.invoke('getUsers', this.searchThis._wfgContext, {
                groupId: results.items[0].id
              })
                .then(users => {
                  results.items = results.items.filter(g => !!users.items.find(u => u.id === this.userId));
                  results.items[0].users = users.items;
                  resolve(results);
                })
                .catch(reject);
            } else {
              resolve(results);
            }
          }.bind(this));
        }.bind({
          userId: filter.users,
          groupId: filter.id,
          searchThis: this
        });

        delete filter.id;
        delete filter.users;
      }

      if (filter.hasOwnProperty('systemIdentifier')) {
        filter.systemIdentifiers = [filter.systemIdentifier];
        additionalProcessing = function (results) {
          if (this.searchThis._searchGroups) {
            return new Promise(function (resolve, reject) {
              let count = 0;
              const callback = () => {
                count++;

                if (count === results.items.length) {
                  resolve(results);
                }
              };

              if (results.items.length <= 0) {
                resolve(results);
                return;
              }

              results.items.forEach(g => {
                this.searchThis._workflowService.invoke('getUsers', this.searchThis._wfgContext, {
                  groupId: g.id
                })
                  .then(users => {
                    g.users = users.items;
                    callback();
                  })
                  .catch(reject);
              });
            }.bind(this));
          }

          return results;
        }.bind({
          systemIdentifier: filter.systemIdentifier,
          searchThis: this
        });

        delete filter.systemIdentifier;
      }
    }

    this._workflowService.invoke(methodName, this._wfgOperationContext, filter)
      .then(additionalProcessing)
      .then(results => {
        const elements = results.items
          .filter(item => item.directoryId === this._directoryId)
          .map(item => {
            const element = new TranslatorClass(TranslatorClass.WFG_TO_SCIM_TYPE, item).translate();

            if (this._searchGroups && element.members) {
              element.members = element.members.map(m => ({
                value: m.id,
                display: m.displayName
              }));
            }

            if (this._attributes) {
              element.includesOnly(...this._attributes);
            } else if (this._excludedAttributes) {
              element.excludes(...this._excludedAttributes);
            }

            return element;
          });

        callback(elements);
      })
      .catch(err => callback(null, err));
  }

  /**
   * Executes the search operation.
   * @return {Promise} A promise indicating success or failure.
   */
  execute () {
    return new Promise(function (resolve, reject) {
      if (
        this._resourceSchema === constants.schemas.user &&
        !this._searchUsers
      ) {
        throw new UnauthorizedSearchError();
      }

      if (
        this._resourceSchema === constants.schemas.group &&
        !this._searchGroups
      ) {
        throw new UnauthorizedSearchError();
      }

      if (
        !this._resourceSchema &&
        !(this._searchUsers && this._searchGroups)
      ) {
        throw new UnauthorizedSearchError();
      }

      this._search((results, err) => {
        if (err) {
          reject(err);
          return;
        }

        resolve(results);
      });
    }.bind(this));
  }
}

module.exports = {
  Search: inject(Search),
  TooManyResultsError,
  SearchError,
  InvalidCountError,
  InvalidFilterError,
  InvalidSortOrderError,
  InvalidStartIndexError,
  UnauthorizedSearchError
};
