/**
 * translator.js
 * Contains the code to translate from and to schmas.
 */
const assert = require('assert');

/**
 * @desc Abstract class for concrete translators.
 * @interface
 */
class Translator {
  static get SCIM_TO_WFG_TYPE () {
    return 'SCIM_TO_WFG_TYPE';
  }

  static get WFG_TO_SCIM_TYPE () {
    return 'WFG_TO_SCIM_TYPE';
  }

  /**
   * @desc Instanciates a new Translator.
   * @param {String} type Type of translator (Which direction to translate)
   * @param {Object} [object=null] The object to translate
   * @constructor
   */
  constructor (type, object = null) {
    assert(type === Translator.SCIM_TO_WFG_TYPE || type === Translator.WFG_TO_SCIM_TYPE);

    this._type = type;
    this._object = object;
  }

  /**
   * @desc Translate the object passed in the constructor.
   * @return {Object} The translated object. `null` if no object was provided in the constructor.
   */
  translate () {
    if (!this._object) {
      return null;
    }

    switch (this._type) {
      case Translator.SCIM_TO_WFG_TYPE:
        return this._translateToWFG(this._object);

      case Translator.WFG_TO_SCIM_TYPE:
        return this._translateToSCIM(this._object);

      default:
        return null;
    }
  }

  /**
   * @desc Translates an attribute name into its translated conterpart.
   * @param {String} attr The name of the attribute to translate.
   * @return {String} The translated attribute.
   */
  translateAttribute (attr) {
    for (let [key, value] of this._getSCIMAttributesMap().entries()) {
      switch (this._type) {
        case Translator.SCIM_TO_WFG_TYPE:
          if (key === attr) {
            return value;
          }
          continue;

        case Translator.WFG_TO_SCIM_TYPE:
          if (value === attr) {
            return key;
          }
          continue;

        default:
          return '';
      }
    }
  }

  /**
   * @desc Retrieve the Map that maps the attributes between two objects.
   * @returns {Map<String, String>} The map of the objects attributes.
   * @abstract
   */
  _getSCIMAttributesMap () {
    throw new Error('Not implemented');
  }

  /**
   * @see {@link translate}
   * @param {Object} object The object to be translated to a WFG object.
   * @return {Object} The translated object.
   * @abstract
   */
  _translateToWFG (object) {
    throw new Error('Not implemented');
  }

  /**
   * @see {@link translate}
   * @param {Object} object The object to be translated to a SCIM object.
   * @return {Object} The translated object.
   * @abstract
   */
  _translateToSCIM (object) {
    throw new Error('Not implemented');
  }
}

module.exports = Translator;
