/**
 * usertranslator.js
 * Contains the definition of the UserTranslator class.
 */
'use strict';
const UserResource = require('../user').UserResource;
const WFGUser = require('../user').WFGUser;
const constants = require('../../models/constants');
const Translator = require('./translator');

const ACCEPTED_PHONE_NUMBER_TYPES = new Set([
  'phone', 'mobile', 'fax', 'pager'
]);

/**
 * Concrete User translator
 */
class UserTranslator extends Translator {
  /**
   * @see {@link Translator#_getSCIMAttributesMap}
   */
  _getSCIMAttributesMap () {
    return new Map([
      ['id', 'id'],
      ['userName', 'userName'],
      ['name', 'lastName'],
      ['name.familyName', 'lastName'],
      ['name.givenName', 'firstName'],
      ['displayName', 'commonName'],
      ['title', 'jobTitle'],
      ['locale', 'defaultLanguage'],
      ['timezone', 'defaultTimeZoneId'],
      ['externalId', 'systemIdentifier'],
      ['active', 'isActive'],
      ['emails.value', 'email'],
      ['phoneNumbers.value', ''],
      ['addresses.streetAddress', 'postalAddress'],
      ['addresses.locality', 'city'],
      ['addresses.region', 'state'],
      ['addresses.postalCode', 'postalCode'],
      ['addresses.country', 'country'],
      ['employeeNumber', 'employeeNumber'],
      ['organization', 'company'],
      ['department', 'department'],
      ['manager', 'managerId']
    ]);
  }

  /**
   * @see {@link Translator#_translateToWFG}
   */
  _translateToWFG (object) {
    let user = new WFGUser();

    user.id = object.id;
    user.userName = object.userName;
    user.systemIdentifier = object.externalId;

    if (typeof object.name === 'object') {
      user.lastName = object.name.familyName;
      user.firstName = object.name.givenName;
    } else {
      user.lastName = object.name;
    }

    user.commonName = object.displayName;
    user.jobTitle = object.title;
    user.defaultLanguage = object.locale;
    user.isActive = object.active;

    if (object.emails && object.emails instanceof Array && object.emails.length > 0) {
      user.email = object.emails[0].value;
    }

    let phoneNumbers = object.phoneNumbers || [];
    for (let number of phoneNumbers) {
      if (ACCEPTED_PHONE_NUMBER_TYPES.has(number.type)) {
        user[number.type] = number.value;
      }
    }

    if (object.addresses && object.addresses instanceof Array && object.addresses.length > 0) {
      const workAddress = object.addresses.find(addr => addr.type === 'work');
      const otherAddress = object.addresses.find(addr => addr.type === 'other');

      if (workAddress) {
        user.postalAddress = workAddress.streetAddress;
        user.city = workAddress.locality;
        user.state = workAddress.region;
        user.postalCode = workAddress.postalCode;
        user.country = workAddress.country;
      }

      if (otherAddress) {
        user.office = otherAddress.formatted;
      }
    }

    if (object.hasOwnProperty(constants.schemas.extensions.user.business)) {
      const businessProperties = object[constants.schemas.extensions.user.business];

      user.employeeNumber = businessProperties.employeeNumber;
      user.company = businessProperties.organization;
      user.department = businessProperties.department;

      if (businessProperties.manager) {
        user.managerId = businessProperties.manager.id;
      }
    }

    Object.cleanProperties(user, false);
    return user;
  }

  /**
   * @see {@link Translator#_translateToSCIM}
   */
  _translateToSCIM (object) {
    let user = new UserResource();

    user.id = object.id;
    user.externalId = object.systemIdentifier;
    user.userName = object.userName;

    if (object.lastName && object.firstName) {
      user.name = {
        givenName: object.firstName,
        familyName: object.lastName
      };
    } else if (object.lastName) {
      user.name = object.lastName;
    }

    user.displayName = object.commonName;
    user.title = object.jobTitle;
    user.locale = object.defaultLanguage;
    user.active = object.isActive;
    user.emails = [];
    user.phoneNumbers = [];
    user.addresses = [];

    if (object.email) {
      user.emails.push({
        value: object.email,
        type: 'work',
        primary: true
      });
    }

    ACCEPTED_PHONE_NUMBER_TYPES.forEach(type => {
      if (object.hasOwnProperty(type) && object[type]) {
        user.phoneNumbers.push({
          value: object[type],
          type
        });
      }
    });

    if (
      object.postalAddress ||
      object.city ||
      object.state ||
      object.postalCode ||
      object.country
    ) {
      user.addresses.push({
        streetAddress: object.postalAddress,
        locality: object.city,
        region: object.state,
        postalCode: object.postalCode,
        country: object.country
      });
    }

    if (object.office) {
      user.addresses.push({
        type: 'other',
        formatted: object.office
      });
    }

    if (
      object.employeeNumber ||
      object.company ||
      object.department ||
      object.managerId
    ) {
      user[constants.schemas.extensions.user.business] = {
        employeeNumber: object.employeeNumber,
        organization: object.company,
        department: object.department,
        manager: {
          id: object.managerId
        }
      };
    }

    Object.cleanProperties(user, false);
    return user;
  }
}

module.exports = UserTranslator;
