/**
 * user.js
 * Constains the definition of a user of the business domain.
 */
'use strict';

const constants = require('../models/constants');
const Resource = require('./resource');

/**
 * Represents a response to send to the user.
 * @class UserResource
 */
class UserResource extends Resource {
  constructor (obj = null) {
    super('User');
    this.schemas = [constants.schemas.user];

    if (obj) {
      for (let prop of Object.getOwnPropertyNames(obj)) {
        this[prop] = obj[prop];
      }
    }
  }

  /**
   * Adds the business schema extension if necessary
   */
  checkAddBusinessExtension () {
    if (
      this.employeeNumber ||
      this.costCenter ||
      this.organization ||
      this.division ||
      this.department ||
      this.manager
    ) {
      this.schemas.push(constants.schemas.extensions.user.business);
    }
  }

  /**
   * @description Adds the display name of the manager to the representation.
   * @param {String} displayName The display name of the manager.
   */
  addManagerInfo (displayName) {
    this.manager.displayName = displayName;
  }

  /**
   * @description Adds the full url location to the representation.
   * @param {String} baseUrl The base url of the resource.
   */
  addUrlLocation (baseUrl) {
    if (this.manager && !Object.isNullOrUndefined(this.manager.id)) {
      this.manager.$ref = `${baseUrl}/${this.manager.id}`;
    }

    this.meta.location = `${baseUrl}/${this.id}`;
  }

  /**
   * @description Format multiple names and titles into a single one.
   * @param {String} honorificPrefix The honorific prefix of the user.
   * @param {String} givenName The given name of the user.
   * @param {String} middleName The middle name of the user.
   * @param {String} lastName The last name of the user.
   * @param {String} honorificSuffix The honorific suffix of the user.
   * @return {String} The formatted name altogether.
   */
  formattedName (
    honorificPrefix = null,
    givenName = null,
    middleName = null,
    lastName = null,
    honorificSuffix = null
  ) {
    let formatted = '';
    let space = ' ';

    if (honorificPrefix) {
      formatted += honorificPrefix + space;
    }

    if (givenName) {
      formatted += givenName + space;
    }

    if (middleName) {
      formatted += middleName + space;
    }

    if (lastName) {
      formatted += lastName + space;
    }

    if (honorificSuffix) {
      formatted += honorificSuffix + space;
    }

    return formatted.substr(0, formatted.length - 1);
  }

  /**
   * @description Format the address from various values.
   * @param {String} [streetAddress=null] The street address.
   * @param {String} [locality=null] The locality.
   * @param {String} [region=null] The region.
   * @param {String} [postalCode=null] The postal code.
   * @param {String} [country=null] The country.
   * @return {String} The formatted address.
   */
  formattedAddress (
    streetAddress = null,
    locality = null,
    region = null,
    postalCode = null,
    country = null
  ) {
    let formatted = '';
    let space = ' ';
    let thirdPartPresent = (region || postalCode || country);

    if (locality || thirdPartPresent) {
      formatted += streetAddress || '\n';
    } else {
      formatted += streetAddress || '';
    }

    if (locality && thirdPartPresent) {
      formatted += '\n' + locality + ', ';
    } else if (locality) {
      formatted += '\n' + locality + space;
    }

    if (region) {
      formatted += region + space;
    }

    if (postalCode) {
      formatted += postalCode + space;
    }

    if (country) {
      formatted += country + space;
    }

    return formatted.substr(0, formatted.length - 1);
  }
}

/**
 * Represents a WorkflowGen user.
 * @class WFGUser
 */
class WFGUser {}

module.exports = {
  UserResource,
  WFGUser
};
