/**
 * server.js
 * Server initialization
 */
'use strict';
const express = require('express');
const bodyParser = require('body-parser');
const cors = require('cors');
const uuid = require('uuid/v1');
const { logger, FileLogger } = require('./utils/altlogger');
const config = require('./config');
const ErrorHelper = require('./models/helpers/error');
const { Actions } = require('./models/helpers/logger');
const { Injector } = require('./utils/injector');
const WorkflowService = require('./services/workflow/workflowservice');
const path = require('path');

require('./utils/extensions'); // Load extensions

const ROOT_URL = '(\\/\\w+)?\\/scim';
const USERS_URL_REGEX = new RegExp(`${ROOT_URL}/Users`);
const GROUPS_URL_REGEX = new RegExp(`${ROOT_URL}/Groups`);
const app = express();

if (process.env.NODE_ENV === 'test' && config.test.runMode !== 'app') {
  require('./test/stub').stubWorkflowService();
} else {
  Injector.configuration = {
    classes: [
      WorkflowService
    ],
    args: [
      [config.workflowService]
    ]
  };
}

FileLogger.configure({
  formatter: (options) => {
    const action = (options.meta && options.meta.action) || Actions.GENERAL;
    const directory = (options.meta && options.meta.directory) || 'DIRECTORY';
    return `${options.timestamp.toJSON()};${directory};${action};${options.message}`;
  },
  getLogFilePath: (meta) => {
    const today = new Date();
    const year = today.getFullYear();
    const month = today.getMonth() + 1;
    const day = today.getDate();
    const formattedDate = `${year}${month.toString().length === 1 ? '0' + month : month}${day.toString().length === 1 ? '0' + day : day}`;

    return path.join(config.app.logsPath, `${formattedDate}_000000_${meta.directory}.log`);
  }
});

logger.console.info('Application NODE_ENV: %s', process.env.NODE_ENV);
logger.console.info(`Workflow service debug enabled: ${config.workflowService.debugEnabled}`);

// Enable Cross Origin Resource Sharing
app.use('*', cors());

// Parse application/json
app.use(bodyParser.json({
  type: 'application/scim+json'
}));

// Catch json syntax error
app.use(function (error, req, res, next) {
  if (error instanceof SyntaxError) {
    logger.console.error('Invalid JSON');
    ErrorHelper.sendInvalidSyntax(res);
    return;
  }

  next();
});

// Authorization middleware
app.use(function (req, res, next) {
  res.set('Content-Type', 'application/scim+json');

  req.uuid = uuid();
  logger.console.info(`New request ${req.uuid}`);

  const authorizationHeader = req.header('authorization'); // e.g. Authorization: Bearer <token>
  let token;

  if (authorizationHeader) {
    const authComponents = authorizationHeader.split(' ');

    if (authComponents.length !== 2) {
      logger.console.error(`Request ${req.uuid} invalid bearer token`);
      ErrorHelper.sendUnauthorizedError(res);
      return;
    }

    if (authComponents[0] !== 'Bearer') {
      logger.console.error(`Request ${req.uuid} invalid bearer token`);
      ErrorHelper.sendUnauthorizedError(res);
      return;
    }

    token = authComponents[1];

    if (token === null || token === '') {
      logger.console.error(`Request ${req.uuid} missing bearer token`);
      ErrorHelper.sendUnauthorizedError(res);
      return;
    }
  } else {
    logger.console.error(`Request ${req.uuid} missing authorization header`);
    ErrorHelper.sendUnauthorizedError(res);
    return;
  }

  req.authToken = token;
  next();
});

// Router configuration
const usersController = require('./controllers/users');
const groupsController = require('./controllers/groups');

app.use(USERS_URL_REGEX, usersController);
app.use(GROUPS_URL_REGEX, groupsController);

// Start Express Server
if (process.env.NODE_ENV !== 'test' || config.test.runMode !== 'app') {
  app.listen(config.app.port, function () {
    logger.console.info(`SCIM server running on port ${config.app.port}`);
  });
}

module.exports = app;
