/**
 * workflowservice.js
 * Workflow Service
 */
'use strict';
const edge = require('edge-js');

// Use bluebird Promise to fix:
// https://github.com/tjanczuk/edge/issues/485
// https://github.com/Microsoft/vscode/issues/1746
const Promise = require('bluebird');

/**
 *  Represents a Workflow Service
 *  @class WorkflowService
 */
class WorkflowService {
  /**
   * Creates a new instance of WorkflowService.
   * @constructor
   * @param {Object} config A service configuration.
   */
  constructor (config) {
    this._config = config;
    this._isStarted = false;
    this._isStarting = false;
    this._isStartError = false;
    this._defaultExceptionStatusCode = 500;
    this._exceptionStatusCodes = {
      'System.Exception': 500,
      'System.Configuration.ConfigurationErrorsException': 500,
      'System.ArgumentException': 400,
      'System.InvalidOperationException': 500,
      'System.IO.FileNotFoundException': 400,
      'System.TimeoutException': 408,
      'System.UnauthorizedAccessException': 403,
      'System.Security.Authentication.AuthenticationException': 401,
      'System.Data.DuplicateNameException': 409
    };

    // Load the workflow service assembly
    this._createWorkflowServiceFacade = edge.func({
      assemblyFile: this._config.serviceAssemblyFilePath,
      typeName: 'Advantys.Workflow.ServiceFacade',
      methodName: 'CreateProxyAsync'
    });
  }

  /**
   * Starts the workflow service.
   * @return {Promise<Object>} A promise.
   */
  start () {
    const self = this;

    self._isStarting = true;

    return new Promise(function (resolve, reject) {
      self._createWorkflowServiceFacade(self._config, function (error, result) {
        if (error) {
          self._isStarted = false;
          self._isStarting = false;
          self._isStartError = true;
          reject(error);
        } else {
          self._workflowFacade = result;
          self._isStarted = true;
          self._isStarting = false;
          self._isStartError = false;
          resolve();
        }
      });
    });
  }

  /**
  * Invokes a workflow operation asynchronously.
  * @param {String} operation An operation to execute.
  * @param {Object} context A workflow context.
  * @param {Object} args Arguments.
  * @return {Promise<Object>} A promise.
  */
  invoke (operation, context, args) {
    return this._invokeOperationWithRetry(operation, context, args, 0);
  }

  /**
  * Invokes a workflow operation asynchronously with retries
  * @param {String} operation An operation to execute.
  * @param {Object} context A workflow context.
  * @param {Object} args Arguments.
  * @param {Integer} retriedCount Current number of retries.
  * @return {Promise<Object>} A promise.
  */
  _invokeOperationWithRetry (operation, context, args, retriedCount) {
    const self = this;

    // The service is ready to invoke operation
    if (self._isStarted) {
      return (self._invokeOperation(operation, context, args));
    }
    // The service is in error
    if (self._isStartError) {
      return Promise.reject(new Error('Workflow Service start failed.'));
    }
    // Start the service if needed
    if (!self._isStarting) {
      self.start()
        .then(result => {
          self._config.logHandler('Workflow Service is started');
        })
        .catch(error => {
          self._config.logHandler('Workflow Service start error: ' + error);
        });
    }
    // Retry the operation
    retriedCount++;
    if (self._config.debugEnabled) {
      self._config.logHandler('Workflow Service: invokeWithRetry: retry count:' + retriedCount);
    }
    if (retriedCount > self._config.maxOperationRetries) {
      self._config.logHandler('Workflow Service: invokeWithRetry: error: Workflow Service is not started: max retries reached');
      var customError = new Error('Workflow Service is not started.');
      customError.statusCode = self._defaultExceptionStatusCode;
      return Promise.reject(customError);
    }

    return Promise.delay(self._config.operationRetryDelay)
      .then(() => {
        return self._invokeOperationWithRetry(operation, context, args, retriedCount);
      });
  }

  /**
   * Invokes a workflow operation asynchronously.
   * @param {String} operation An operation to execute.
   * @param {Object} context A workflow context.
   * @param {Object} args Arguments.
   * @return {Promise<Object>} A promise.
   */
  _invokeOperation (operation, context, args) {
    const self = this;

    return new Promise(function (resolve, reject) {
      if (context.operationCount === self._config.maxOperationCount) {
        reject(new Error('The maximum number of operation per request has been reached.'));
        return;
      }
      context.operationCount++;
      var operationContext = {};
      if (context.authUserName) {
        operationContext['authUserName'] = context.authUserName;
      }
      if (context.impersonateUserName) {
        operationContext['impersonateUserName'] = context.impersonateUserName;
      }
      if (context.hookToken) {
        operationContext['hookToken'] = context.hookToken;
      }
      if (context.scimToken) {
        operationContext['scimToken'] = context.scimToken;
      }
      if (context.appClientId) {
        operationContext['appClientId'] = context.appClientId;
      }
      self._workflowFacade.invoke(
        {
          operation: operation,
          context: operationContext,
          args: args
        },
        function (error, result) {
          if (error) {
            if (self._config.debugEnabled) {
              self._config.logHandler('Workflow Service: invokeOperation: error: ' +
                error.message + ' Service stack: ' + error.StackTrace);
            }
            context.operationErrorCount++;
            error.statusCode = self._exceptionStatusCodes[error.name]
              ? self._exceptionStatusCodes[error.name]
              : self._defaultExceptionStatusCode;
            reject(error);
          } else {
            resolve(result);
          }
        });
    });
  }
}

module.exports = WorkflowService;
