/**
 * altlogger.js
 * Contains the implementation of the Logger class.
 */
'use strict';
const path = require('path');
const print = require('format');
const lockfile = require('proper-lockfile');
const fs = require('fs');
const { execSync } = require('child_process');

/**
 * The option object passed to some logger methods.
 * @typedef {object} Options
 * @property {string} Options.level The level of the message.
 * @property {string} Options.message The message to log.
 * @property {object} [Options.meta] The metadata parsed from the arguments.
 */

/**
 * @name GetLogFilePathFunc
 * @function
 * @param {object} [meta] The metadata passed in the arguments.
 * @returns {string} The path to the log file.
 */

/**
 * Represents an abstract logger.
 */
class Logger {
  /**
   * Parses the args and builds an object to passe it to the handlers.
   *
   * The first argument can be either a Date representing the timetamp or a
   * String representing the level. If the first argument is a Date, the second
   * must be a String representing the level.
   *
   * Le last argument can be an Object and is optional. This object is whatever
   * you want to pass to the handlers to format or output.
   *
   * The argument after the level is the log message or the format of the log
   * message as a String. The other arguments after that are the corresponding
   * variables that you want to format in the format string.
   *
   * This function uses the unix format string in the likes of printf and sprintf.
   * @param {...*} args arguments to pass to the handlers.
   * @returns {Promise<void>} A promise for asynchronous loggers.
   */
  _log (...args) {
    let options = {};

    if (args.length > 0) {
      if (args[0] instanceof Date) {
        options.timestamp = args[0];
        options.level = args[1];
        args.splice(0, 2);
      } else {
        options.level = args[0];
        options.timestamp = new Date();
        args.splice(0, 1);
      }

      if (typeof args[args.length - 1] === 'object') {
        options.meta = args[args.length - 1];
        args.pop();
      }
    }

    if (args.length > 1) {
      const format = args[0];

      args.splice(0, 1);
      options.message = print(format, ...args);
    } else if (args.length > 0) {
      options.message = args[0];
    } else {
      options.message = '';
    }

    return this._handleOutput(this._formatter(options), options);
  }

  /**
   * Method that handles the writing of the log message.
   *
   * You can use the value returned by the promise to do additionnal processing
   * if you configured the output correctly.
   * @param {string} formattedText The string returned by the formatter.
   * @param {Options} options The options parsed from the arguments.
   * @returns {Promise<any>} A promise that returns what's was configured
   * @abstract
   * @protected
   */
  _handleOutput (formattedText, options) {
    throw new Error('Not implemented');
  }

  /**
   * Method that handles the format of the message.
   * @param {Options} options The options parsed from the arguments.
   * @returns {string} The formatted message to log.
   * @abstract
   * @protected
   */
  _formatter (options) {
    throw new Error('Not implemented');
  }

  /**
   * Logs an error message.
   * @see {@link Logger#_log} for precision on args and return value.
   */
  error (...args) {
    return this._log('error', ...args);
  }

  /**
   * Logs a warning message.
   * @see {@link Logger#_log} for precision on args and return value.
   */
  warn (...args) {
    return this._log('warn', ...args);
  }

  /**
   * Logs an info message.
   * @see {@link Logger#_log} for precision on args and return value.
   */
  info (...args) {
    return this._log('info', ...args);
  }

  /**
   * Logs a debug message.
   * @see {@link Logger#_log} for precision on args and return value.
   */
  debug (...args) {
    return this._log('debug', ...args);
  }

  /**
   * Configures the logger according to the app' specific needs.
   * @param {object} options The configurable options for the logger.
   * @param {function} [options.formatter] A function that formats the message to log.
   */
  static configure (options) {
    if (options.formatter && typeof options.formatter === 'function') {
      this.prototype._formatter = options.formatter;
    }
  }
}

/**
 * Concrete logger that outputs to the console.
 */
class ConsoleLogger extends Logger {
  /**
   * @override
   * @inheritDoc
   */
  _handleOutput (formattedText) {
    console.log(formattedText);
    return Promise.resolve();
  }

  /**
   * @override
   * @inheritDoc
   */
  _formatter (options) {
    return `${options.timestamp.toJSON()} ${(options.level && options.level.toUpperCase()) || 'INFO'} ${options.message}`;
  }
}

class FileLogger extends Logger {
  /**
   * @override
   * @inheritDoc
   */
  _log (...args) {
    const timestamp = new Date();
    let logFilePath;

    if (args.length > 0 && typeof args[args.length - 1] === 'object') {
      const meta = args[args.length - 1];
      logFilePath = this._getLogFilePath(meta);
    } else {
      logFilePath = this._getLogFilePath({});
    }

    const logsPath = path.dirname(logFilePath);

    if (!fs.existsSync(logsPath)) {
      execSync(`mkdir -p ${logsPath}`);
    }

    if (!fs.existsSync(logFilePath)) {
      fs.writeFileSync(logFilePath, '');
    }

    return new Promise((resolve, reject) => {
      lockfile.lock(logFilePath, {
        retries: {
          retries: 10,
          minTimeout: 500
        }
      }, (err) => {
        if (err) {
          reject(err);
          return;
        }

        super._log(...[timestamp].concat(args))
          .then(() => {
            lockfile.unlockSync(logFilePath);
            resolve();
          })
          .catch(reject);
      });
    });
  }

  /**
   * @override
   * @inheritDoc
   */
  _handleOutput (formattedText, options) {
    return new Promise((resolve, reject) => {
      fs.appendFile(this._getLogFilePath(options.meta), '\n' + formattedText, (err) => {
        if (err) {
          reject(err);
          return;
        }

        resolve();
      });
    });
  }

  /**
   * @override
   * @inheritDoc
   */
  _formatter (options) {
    return `${options.timestamp.toJSON()} ${(options.level && options.level.toUpperCase()) || 'INFO'} ${options.message}`;
  }

  /**
   * Get the default log file path. Configurable.
   * @see {@link FileLogger#configure} for how to configure.
   * @private
   * @param {object} [meta] The metadata parsed in the arguments.
   */
  _getLogFilePath (meta) {
    const today = new Date();
    const year = today.getFullYear();
    const month = today.getMonth() + 1;
    const day = today.getDate();
    const formattedDate = `${year}${month.toString().length === 1 ? '0' + month : month}${day.toString().length === 1 ? '0' + day : day}`;

    return path.join(path.normalize('./logs'), `${formattedDate}.log`);
  }

  /**
   * @override
   * @see {@link Logger#configure} for more options.
   * @param {object} options The configurable options.
   * @param {GetLogFilePathFunc} [options.getLogFilePath] A function that returns the
   */
  static configure (options) {
    super.configure(options);

    if (options.getLogFilePath && typeof options.getLogFilePath === 'function') {
      this.prototype._getLogFilePath = options.getLogFilePath;
    }
  }
}

module.exports = {
  logger: {
    console: new ConsoleLogger(),
    file: new FileLogger()
  },
  ConsoleLogger,
  FileLogger
};
