/**
 * expressions.js
 * Typed expressions for filtering.
 */
'use strict';

const Operators = require('./operators');

/**
 * General expression class.
 * @class Expression
 */
class Expression {
  /**
   * Creates an instance of Expression.
   * @param {Expression} left The left expression.
   * @param {Expression} right The right expression.
   * @param {String} op The operand to apply between the expressions.
   */
  constructor (left, right, op) {
    this.left = left;
    this.right = right;
    this.operator = op;
  }

  // /**
  //  * Returns an SQL string that corresponds to the expression tree.
  //  * @param {Map} [attributesMap=null] Maps the attribute to their value in the SQL statement.
  //  * @return {String} The where clause generated.
  //  */
  // generateSqlWhere(attributesMap = null) {
  //   let sql = Squel.expr();

  //   let generatedSql = this.getSql(sql, attributesMap);
  //   let params = typeof generatedSql.clause === 'undefined' ?
  //     [generatedSql] : [generatedSql.clause, generatedSql.param];

  //   return sql.and(...params).toString();
  // }

  /**
   * Applies the expression to an object.
   * @param {Object} obj The object to apply the expression to.
   * @return {Object} The filtered object.
   */
  apply (obj) {
    return this.test(obj);
  }
}

/**
 * A compound expression. Left and right operand can be other expressions.
 * @class CompoundExpression
 * @extends {Expression}
 */
class CompoundExpression extends Expression {
  /**
   * Tests a compound expression tree.
   * @param {Object} elem The element to apply boolean operators on.
   * @return {Boolean} The boolean result.
   */
  test (elem) {
    if (Operators.GroupingOperator.isTokenGroupingOperator(this.operator)) {
      let leftResult = this.left ? this.left.test() : null;

      return Operators.GroupingOperator.getBooleanFunction(
        leftResult,
        this.right.test(elem),
        this.operator
      );
    } else if (Operators.LogicalOperator.isLogicalOperator(this.operator)) {
      return Operators.LogicalOperator.getBooleanFunction(
        this.left.test(elem),
        this.right.test(elem),
        this.operator
      );
    }
  }

  // /**
  //  * Generates the sql expression for the current compound expression.
  //  * @param {any} sql The current sql string or tree.
  //  * @param {Map} attributesMap Map to replace attribute names.
  //  * @return {any} The SQL string or expression tree.
  //  */
  // getSql(sql, attributesMap = null) {
  //   let hasOperatorNot = Operators
  //     .GroupingOperator.isTokenGroupingOperator(this.operator);
  //   let innerExpr = Squel.expr();
  //   let leftResult;

  //   if (!hasOperatorNot) {
  //     leftResult = this.left.getSql(innerExpr, attributesMap);
  //   }
  //   let rightResult = this.right.getSql(innerExpr, attributesMap);

  //   // If operator is a not, simply return it as a string.
  //   if (hasOperatorNot) {
  //     let params = rightResult.clause ?
  //       [rightResult.clause, rightResult.param] : [rightResult];
  //     return 'NOT (' + Squel.expr().and(...params) + ')';
  //   }

  //   // Else return an inner SQL expression tree.
  //   return Operators.LogicalOperator.getSqlWhereClause(
  //     leftResult,
  //     rightResult,
  //     this.operator,
  //     innerExpr
  //   );
  // }
}

/**
 * A simple expression. Left and right operand are attributes.
 * @class SimpleExpression
 * @extends {Expression}
 */
class SimpleExpression extends Expression {
  /**
   * Tests a simple expression.
   * @param {Object} elem The element to apply tests on.
   * @return {Boolean} The boolean result.
   */
  test (elem) {
    return Operators.AttributeOperator.getBooleanFunction(
      this.left,
      this.operator,
      this.right,
      elem
    );
  }

  // /**
  //  * Generates a simple attribute comparison in SQL.
  //  * @param {String} sql The current SQL tree.
  //  * @param {Map} attributesMap Map to replace attribute names.
  //  * @return {Object} The clause and parameters of the attribute comparison.
  //  */
  // getSql(sql, attributesMap = null) {
  //   let leftOperand = this.left;
  //   let rightOperand = this.right;

  //   if (attributesMap) {
  //     leftOperand = attributesMap.get(leftOperand) || leftOperand;
  //   }

  //   switch (this.operator) {
  //     case 'SW':
  //       rightOperand += '%';
  //       break;
  //     case 'EW':
  //       rightOperand = '%' + rightOperand;
  //       break;
  //     case 'CO':
  //       rightOperand = '%' + rightOperand + '%';
  //       break;
  //     default:
  //       break;
  //   }

  //   return {
  //     clause: Operators.AttributeOperator.getSqlWhereClause(
  //       leftOperand,
  //       this.operator
  //     ),
  //     param: rightOperand
  //   };
  // }
}

module.exports = {
  Expression,
  SimpleExpression,
  CompoundExpression
};
