/**
 * lexer.js
 * Used to tokenize strings.
 */
'use strict';

const Operators = require('./operators');

/**
 * Tokenizes an input string.
 * @class FilterLexer
 */
class FilterLexer {
  /**
   * Creates an instance of FilterLexer.
   * @param {String|Array} filter The string to tokenize, or an array of tokens.
   */
  constructor (filter) {
    if (filter instanceof Array) {
      this.tokens = filter;
    } else {
      this.filter = filter;

      if (filter && filter !== '') {
        let cleanFilter = filter;
        // Add space between parenthesis
        cleanFilter = cleanFilter.replace(/\(/g, ' ( ');
        cleanFilter = cleanFilter.replace(/\)/g, ' ) ');

        // Convert multiple spaces to 1
        cleanFilter = cleanFilter.replace(/ +/g, ' ');

        // Eliminate the space between NOT and ( and capitalize NOT
        cleanFilter = cleanFilter.replace(/not \(/g, 'NOT(');

        // Trim spaces
        cleanFilter = cleanFilter.trim();

        // Tokenize
        let tokens = cleanFilter.split(' ');

        for (let i = 0; i < tokens.length; i++) {
          let token = tokens[i];

          if (
            Operators.AttributeOperator.isAttributeOperator(token) ||
            Operators.LogicalOperator.isLogicalOperator(token)
          ) {
            tokens[i] = token.toUpperCase();
          }
        }

        // Regroup value strings
        for (let i = 0; i < tokens.length; i++) {
          if (tokens[i].startsWith('"')) {
            while (!tokens[i].endsWith('"')) {
              tokens[i] = tokens[i] + ' ' + tokens[i + 1];
              tokens.splice(i + 1, 1);
            }
          }
        }

        this.tokens = tokens;
      }
    }
  }

  /**
   * Provides the next token.
   * @return {String} The token.
   */
  next () {
    let next = this.tokens.shift();
    this.previousToken = next;
    return next;
  }

  /**
   * Returns the last token consumed.
   * @return {String} The token.
   */
  previous () {
    return this.previousToken;
  }

  /**
   * Peeks at the next token.
   * @return {String} The token.
   */
  peek () {
    return this.tokens[0];
  }

  /**
   * Verifies if a token is available.
   * @return {Boolean} If a token is available.
   */
  hasNext () {
    return this.tokens.length > 0;
  }
}

module.exports = FilterLexer;
