/**
 * operators.js
 * Typed operators for filtering.
 */
'use strict';

const constants = require('./constants');

class UnsupportedOperationError extends Error {
  constructor () {
    super('This filter operation is not supported');
  }
}

/**
 * Operator that tests a value.
 * @class AttributeOperator
 */
class AttributeOperator {
  /**
   * Checks if the token is an attribute operator.
   * @static
   * @param {String} token The token to check.
   * @return {Boolean} If the token is an attribute operator.
   */
  static isAttributeOperator (token) {
    return constants
      .filterOperators.attributeOperators.includes(token.toUpperCase());
  }

  /**
   * Comparison tests between two values.
   * @static
   * @param {String} attribute The atribute to check on object.
   * @param {String} operator The comparison operator to use.
   * @param {String} valueToTest The value to test.
   * @param {Object} object The object to test.
   * @return {Boolean} If the valueToTest matches the object's value
   */
  static getBooleanFunction (attribute, operator, valueToTest, object) {
    let value = object[attribute];

    switch (operator) {
      case 'EQ':
        return String(value) === valueToTest;
      case 'NE':
        return String(value) !== valueToTest;
      case 'CO':
        if (
          !(typeof value === 'string') || !(typeof valueToTest === 'string')
        ) {
          throw new UnsupportedOperationError();
        }
        return value.includes(valueToTest);
      case 'SW':
        if (
          !(typeof value === 'string') || !(typeof valueToTest === 'string')
        ) {
          throw new UnsupportedOperationError();
        }
        return value.startsWith(valueToTest);
      case 'EW':
        if (
          !(typeof value === 'string') || !(typeof valueToTest === 'string')
        ) {
          throw new UnsupportedOperationError();
        }
        return value.endsWith(valueToTest);
      case 'PR':
        return typeof value !== 'undefined';
      case 'GT':
        if (typeof value === 'boolean') {
          throw new UnsupportedOperationError();
        }
        return value > valueToTest;
      case 'GE':
        if (typeof value === 'boolean') {
          throw new UnsupportedOperationError();
        }
        return value >= valueToTest;
      case 'LT':
        if (typeof value === 'boolean') {
          throw new UnsupportedOperationError();
        }
        return value < valueToTest;
      case 'LE':
        if (typeof value === 'boolean') {
          throw new UnsupportedOperationError();
        }
        return value <= valueToTest;
      default:
        throw new Error('Unsupported attribute operator');
    }
  }

  /**
   * Generates a simple attribute comparison test.
   * @static
   * @param {String} attribute The attribute to operate on.
   * @param {String} operator The operator.
   * @return {String} The test.
   */
  static getSqlWhereClause (attribute, operator) {
    switch (operator) {
      case 'EQ':
        return `${attribute} = ?`;
      case 'NE':
        return `${attribute} <> ?`;
      case 'CO':
      case 'SW':
      case 'EW':
        return `${attribute} LIKE ?`;
      case 'PR':
        return `${attribute} IS NOT NULL`;
      case 'GT':
        return `${attribute} > ?`;
      case 'GE':
        return `${attribute} >= ?`;
      case 'LT':
        return `${attribute} < ?`;
      case 'LE':
        return `${attribute} <= ?`;
      default:
        throw new Error('Unsupported attribute operator');
    }
  }
}

/**
 * Operator that performs logical tests.
 * @class LogicalOperator
 */
class LogicalOperator {
  /**
   * Checks if the token is a logical operator.
   * @static
   * @param {String} token The token to check.
   * @return {Boolean} If the token is an logical operator.
   */
  static isLogicalOperator (token) {
    return constants
      .filterOperators.logicalOperators.includes(token.toUpperCase());
  }

  /**
   * Logical test between two booleans.
   * @static
   * @param {Boolean} left The left boolean.
   * @param {Boolean} right The right boolean.
   * @param {String} operator The operator to use.
   * @return {Boolean} Test result between left and right.
   */
  static getBooleanFunction (left, right, operator) {
    switch (operator) {
      case 'AND':
        return left && right;
      case 'OR':
        return left || right;
      default:
        throw new Error('Unsupported logical operator');
    }
  }

  /**
   * Generates a logical SQL expression tree.
   * @static
   * @param {Object} left The left clause.
   * @param {Object} right The right clause.
   * @param {String} operator The logical operator.
   * @param {Object} sql The current SQL expression tree.
   * @return {Object} The modified SQL expression tree.
   */
  static getSqlWhereClause (left, right, operator, sql) {
    let leftIsSimple = typeof left.clause !== 'undefined';
    let rightIsSimple = typeof right.clause !== 'undefined';
    let leftParams = leftIsSimple ? [left.clause, left.param] : [left];
    let rightParams = rightIsSimple ? [right.clause, right.param] : [right];

    switch (operator) {
      case 'AND':
        return sql.and(...leftParams).and(...rightParams);
      case 'OR':
        return sql.or(...leftParams).or(...rightParams);
      default:
        throw new Error('Unsupported logical operator');
    }
  }
}

/**
 * Operator that groups multiples expressions.
 * @class GroupingOperator
 */
class GroupingOperator {
  /**
   * Checks if the symbol is a grouping operator.
   * @static
   * @param {String} symbolToTest The symbol to check.
   * @return {Boolean} If the symbol is an attribute operator.
   */
  static isSymbolGroupingOperator (symbolToTest) {
    let operators = constants.filterOperators.groupingOperators;

    for (let op in operators) {
      if (operators[op] === symbolToTest) {
        return true;
      }
    }

    return false;
  }

  /**
   * Checks if the token is a grouping operator.
   * @static
   * @param {String} token The symbol to check.
   * @return {Boolean} If the token is an attribute operator.
   */
  static isTokenGroupingOperator (token) {
    return Boolean(constants.filterOperators.groupingOperators[token]);
  }

  /**
   * Returns the grouping operator from a symbol.
   * @static
   * @param {String} symbolToFind The symbol to find the operator from.
   * @return {String} The operator.
   */
  static fromSymbol (symbolToFind) {
    let operators = constants.filterOperators.groupingOperators;

    for (let op in operators) {
      if (operators[op] === symbolToFind) {
        return op;
      }
    }
  }

  /**
   * Tests a grouping operator (i.e. not ()).
   * @static
   * @param {Boolean} left The left boolean.
   * @param {Boolean} right The right boolean.
   * @param {String} operator The operator to use.
   * @return {Boolean} The grouping result.
   */
  static getBooleanFunction (left, right, operator) {
    switch (operator) {
      case 'NP':
        return !right;
      default:
        throw new Error('Unsupported logical operator');
    }
  }
}

module.exports = {
  AttributeOperator,
  LogicalOperator,
  GroupingOperator,
  UnsupportedOperationError
};
