/**
 * parser.js
 * Parses tokenized arrays.
 */
'use strict';

const Operators = require('./operators');
const Expressions = require('./expressions');
const FilterLexer = require('./lexer');

class InvalidOperatorError extends Error {}
class InvalidAttributeOperatorError extends InvalidOperatorError {}

/**
 * Parses a list of tokens.
 * @class FilterParser
 */
class FilterParser {
  constructor (lexer) {
    this.lexer = lexer;
  }

  static get Errors () {
    return {
      InvalidOperatorError,
      InvalidAttributeOperatorError
    };
  }

  /**
   * Parse the input tokens.
   * @return {Expression} The parsed expression.
   */
  parse () {
    return this.leftToRightParse(null);
  }

  /**
   * Parses the tokens from left to right.
   * @param {Expression} expressionIn The current expression.
   * @return {Expression} The next expression.
   */
  leftToRightParse (expressionIn) {
    let expressionOut = expressionIn;
    while (this.lexer.hasNext()) {
      expressionOut = this.recursiveParse(expressionOut);
    }
    return expressionOut;
  }

  /**
   * Recursively parses an expression.
   * @param {Expression} expressionIn The expression to parse.
   * @return {Expression} The next expression.
   */
  recursiveParse (expressionIn) {
    let expressionOut = expressionIn;

    if (this.lexer.hasNext()) {
      let token = this.lexer.next();

      if (this.isAttributeName(token)) {
        let attributeExpression = this.parseAttributeExpression(token);
        expressionOut = this.recursiveParse(attributeExpression);
      } else if (Operators.LogicalOperator.isLogicalOperator(token)) {
        let logicalExpression = this.parseCompoundExpression(
          expressionIn,
          token
        );

        expressionOut = this.recursiveParse(logicalExpression);
      } else if (Operators.GroupingOperator.isSymbolGroupingOperator(token)) {
        let groupingOperator = Operators.GroupingOperator.fromSymbol(token);

        if (groupingOperator === 'OP') {
          expressionOut = this.parseGroupedExpression(false);
        } else if (groupingOperator === 'NP') {
          expressionOut = this.parseGroupedExpression(true);
        } else if (groupingOperator === 'CP') {
          return expressionIn;
        }
      }

      return expressionOut;
    }

    return expressionIn;
  }

  /**
   * Parses a simple expression between two operands.
   * @param {String} attributeName The attribute to operate on.
   * @return {SimpleExpression} A simple expression between two operands.
   */
  parseAttributeExpression (attributeName) {
    let left = attributeName;
    let operator = this.lexer.next();
    let right;

    if (!Operators.AttributeOperator.isAttributeOperator(operator)) {
      throw new InvalidAttributeOperatorError(`Operator "${operator}" is not valid.`);
    }

    if (operator !== 'PR') {
      right = this.lexer.next();
      right = right.substring(1, right.length - 1);
    }

    return new Expressions.SimpleExpression(left, right, operator);
  }

  /**
   * Parses a compound expression between two expressions.
   * @param {Expression} left The left expression.
   * @param {String} operator The operator.
   * @return {CompoundExpression} A compound expression between two expressions.
   */
  parseCompoundExpression (left, operator) {
    let right = this.leftToRightParse(null);

    return new Expressions.CompoundExpression(left, right, operator);
  }

  /**
   * Parses a group of expressions.
   * @param {Boolean} negativeGroup If the group is a not () operator.
   * @return {Expression} A grouped expression.
   */
  parseGroupedExpression (negativeGroup) {
    let expressionOut;
    let tokens = [];
    let token;

    do {
      token = this.lexer.next();
      tokens.push(token);
    } while (token !== ')');

    let previousLexer = this.lexer;
    this.lexer = new FilterLexer(tokens);
    let groupedExpression = this.leftToRightParse(null);
    this.lexer = previousLexer;

    if (negativeGroup) {
      let operator = 'NP';
      let right = groupedExpression;
      expressionOut = new Expressions.CompoundExpression(null, right, operator);
    } else {
      expressionOut = groupedExpression;
    }

    return expressionOut;
  }

  /**
   * Determines if a token is an attribute.
   * @param {String} token The token to check.
   * @return {Boolean} If the token is an attribute.
   */
  isAttributeName (token) {
    return (!token.startsWith('\'') ||
      !token.endsWith('\'')) &&
      !this.isOperator(token);
  }

  /**
   * Determines if a token is an operator.
   * @param {String} token The token to check.
   * @return {Boolean} If the token is an operator.
   */
  isOperator (token) {
    return Operators.AttributeOperator.isAttributeOperator(token) ||
      Operators.GroupingOperator.isSymbolGroupingOperator(token) ||
      Operators.LogicalOperator.isLogicalOperator(token);
  }
}

module.exports = FilterParser;
