/**
 * injector.js
 * Contains the definition of the dependency injector.
 */
'use strict';

const getClassInjectRegex = (className) => new RegExp(`\\/\\*\\s?${className}\\s?\\*\\/`);

/**
 * @desc Object dependency injector. Only holds the configuration for the injection.
 * It has to be used in conjunction with the injection decorator.
 */
class Injector {
  /**
   * @desc Configuration setter for the injection.
   * @param {Object} value The configuration object.
   * @param {Array<Object>} value.classes Classes to be injected into other objects.
   * @param {Array<Array<Object>>} value.args Arguments to be passed to classes when instanciated.
   */
  static set configuration (value) {
    Injector.classes = value.classes && value.classes instanceof Array ? value.classes : [];
    Injector.args = value.args && value.args instanceof Array ? value.args : [];
  }

  /**
   * @desc Get the complete set of arguments to be passed to the injectee class.
   * @param {Array<Object>} passedArguments Arguments passed to the injectee class.
   * @return {Array<Object>} An array containing the passed arguments + the injected classes.
   */
  static _getArgumentsToPass (passedArguments) {
    let injected = [];

    for (let i = 0; i < Injector.classes.length; i++) {
      let injectedArguments = [];

      if (i < Injector.args.length) {
        injectedArguments = Injector.args[i];
      }

      injected.push(new Injector.classes[i](...injectedArguments));
    }

    return passedArguments.concat(injected);
  }

  /**
   * @desc Instanciates an injected class be its name.
   * @param {String} className The name of the classe to be instanciated.
   * @return {Object} The injected object.
   */
  static _instanciateClassToInject (className) {
    const classIndex = Injector.classes
      .map(Class => Class.name)
      .indexOf(className);
    const Class = Injector.classes[classIndex];
    const classArguments = Injector.args[classIndex];

    if (classArguments) {
      return new Class(...classArguments);
    }

    return new Class();
  }
}

/**
 * @desc Decorates a class with injected classes.
 * @param {Class} Wrapped The class to be decorated.
 * @return {Class} The decorated class.
 */
function inject (Wrapped) {
  let wrappedClass = function () {
    const passedArguments = Array.prototype.slice.call(arguments);
    const argumentsToPass = Injector._getArgumentsToPass(passedArguments);

    return new Wrapped(...argumentsToPass);
  };
  const classNameRegexes = Injector.classes
    .map(Class => Class.name)
    .map(getClassInjectRegex);

  Object.getOwnPropertyNames(Wrapped)
    .filter(name =>
      name !== 'constructor' &&
      name !== 'name' &&
      name !== 'length' &&
      name !== 'prototype'
    )
    .forEach(name => {
      // For static functions, we need to check if it needs to have an object
      // injected in its arguments.
      if (typeof Wrapped[name] === 'function') {
        const signature = Wrapped[name].toString().split('\\n')[0];
        const wrappedStaticFunc = function () {
          const passedArguments = Array.prototype.slice.call(arguments);
          let classesToInject = [];

          // The function indicates in the signature a comment that contains the exact
          // name of the class it wants injected.
          classNameRegexes.forEach(regex => {
            const matchedclasses = regex.exec(signature);

            if (matchedclasses && matchedclasses.length > 0) {
              const matchedClass = matchedclasses[0];
              const cleanMatchedClass = matchedClass.replace(/[/*\s?]|[*/\s?]/g, '');
              classesToInject.push(Injector._instanciateClassToInject(cleanMatchedClass));
            }
          });

          return Wrapped[name](...passedArguments.concat(classesToInject));
        };

        wrappedClass[name] = wrappedStaticFunc;
        return;
      }

      wrappedClass[name] = Wrapped[name];
    });
  return wrappedClass;
}

module.exports = {
  Injector,
  inject
};
